
#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
Ground-State DMRG
using pyscf and block2.

Author: Huanchen Zhai, Dec 1, 2020
"""

from block2 import SU2, SZ, Global, OpNamesSet, Threading, ThreadingTypes
from block2 import init_memory, release_memory, set_mkl_num_threads, SiteIndex
from block2 import VectorUInt8, PointGroup, FCIDUMP, QCTypes, SeqTypes, OpNames, Random
from block2 import VectorUBond, VectorDouble, NoiseTypes, DecompositionTypes, EquationTypes
import time
import numpy as np

# Set spin-adapted or non-spin-adapted here
# SpinLabel = SU2
SpinLabel = SZ

if SpinLabel == SU2:
    from block2.su2 import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.su2 import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO
    from block2.su2 import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
else:
    from block2.sz import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.sz import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO
    from block2.sz import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect

try:
    if SpinLabel == SU2:
        from block2.su2 import MPICommunicator
    else:
        from block2.sz import MPICommunicator
    MPI = MPICommunicator()
    from mpi4py import MPI as PYMPI
    comm = PYMPI.COMM_WORLD

    def _print(*args, **kwargs):
        if MPI.rank == 0:
            print(*args, **kwargs)
except ImportError:
    MPI = None
    _print = print


class GSDMRGError(Exception):
    pass


class GSDMRG:
    """
    Ground-State DMRG for molecules.
    """

    def __init__(self, scratch='./nodex', memory=1 * 1E9, omp_threads=8, verbose=2,
                 print_statistics=True, mpi=None, dctr=True):
        """
        Memory is in bytes.
        verbose = 0 (quiet), 2 (per sweep), 3 (per iteration)
        """

        if mpi is not None:
            memory = memory / mpi.size
            if mpi.rank != 0:
                verbose = 0
                print_statistics = False

        Random.rand_seed(0)
        init_memory(isize=int(memory * 0.1),
                    dsize=int(memory * 0.9), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global, omp_threads, omp_threads, 1)
        Global.threading.seq_type = SeqTypes.Simple
        self.fcidump = None
        self.hamil = None
        self.verbose = verbose
        self.scratch = scratch
        self.mpo_orig = None
        self.print_statistics = print_statistics
        self.mpi = mpi
        self.dctr = dctr

        if mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelRuleQC, ParallelRuleNPDMQC
            else:
                from block2.sz import ParallelRuleQC, ParallelRuleNPDMQC
            self.prule = ParallelRuleQC(mpi)
            self.pdmrule = ParallelRuleNPDMQC(mpi)
        else:
            self.prule = None
            self.pdmrule = None

    def init_hamiltonian_fcidump(self, pg, filename):
        """Read integrals from FCIDUMP file.
        pg : point group, pg = 'c1' or 'd2h'
        filename : FCIDUMP filename
        """
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        if self.mpo_orig is not None:
            self.mpo_orig.deallocate()
        self.fcidump = FCIDUMP()
        self.fcidump.read(filename)
        self.orb_sym = VectorUInt8(
            map(PointGroup.swap_d2h, self.fcidump.orb_sym))

        vacuum = SpinLabel(0)
        self.target = SpinLabel(self.fcidump.n_elec, self.fcidump.twos,
                                PointGroup.swap_d2h(self.fcidump.isym))
        self.n_sites = self.fcidump.n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)
        assert pg in ["d2h", "c1"]

    def init_hamiltonian(self, pg, n_sites, n_elec, twos, isym, orb_sym,
                         e_core, h1e, g2e, tol=1E-13, save_fcidump=None):
        """Initialize integrals using h1e, g2e, etc."""
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        if self.mpo_orig is not None:
            self.mpo_orig.deallocate()
        self.fcidump = FCIDUMP()
        if not isinstance(h1e, tuple):
            mh1e = np.zeros((n_sites * (n_sites + 1) // 2))
            k = 0
            for i in range(0, n_sites):
                for j in range(0, i + 1):
                    assert abs(h1e[i, j] - h1e[j, i]) < tol
                    mh1e[k] = h1e[i, j]
                    k += 1
            mg2e = g2e.flatten()
            mh1e[np.abs(mh1e) < tol] = 0.0
            mg2e[np.abs(mg2e) < tol] = 0.0
            self.fcidump.initialize_su2(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        else:
            assert SpinLabel == SZ
            assert isinstance(h1e, tuple) and len(h1e) == 2
            assert isinstance(g2e, tuple) and len(g2e) == 3
            mh1e_a = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e_b = np.zeros((n_sites * (n_sites + 1) // 2))
            mh1e = (mh1e_a, mh1e_b)
            for xmh1e, xh1e in zip(mh1e, h1e):
                k = 0
                for i in range(0, n_sites):
                    for j in range(0, i + 1):
                        assert abs(xh1e[i, j] - xh1e[j, i]) < tol
                        xmh1e[k] = xh1e[i, j]
                        k += 1
                xmh1e[np.abs(xmh1e) < tol] = 0.0
            mg2e = tuple(xg2e.flatten() for xg2e in g2e)
            for xmg2e in mg2e:
                xmg2e[np.abs(xmg2e) < tol] = 0.0
            self.fcidump.initialize_sz(
                n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
        self.orb_sym = VectorUInt8(map(PointGroup.swap_d2h, orb_sym))

        vacuum = SpinLabel(0)
        self.target = SpinLabel(n_elec, twos, PointGroup.swap_d2h(isym))
        self.n_sites = n_sites

        self.hamil = HamiltonianQC(
            vacuum, self.n_sites, self.orb_sym, self.fcidump)

        if save_fcidump is not None:
            self.fcidump.orb_sym = VectorUInt8(orb_sym)
            self.fcidump.write(save_fcidump)
        assert pg in ["d2h", "c1"]

    @staticmethod
    def fmt_size(i, suffix='B'):
        if i < 1000:
            return "%d %s" % (i, suffix)
        else:
            a = 1024
            for pf in "KMGTPEZY":
                p = 2
                for k in [10, 100, 1000]:
                    if i < k * a:
                        return "%%.%df %%s%%s" % p % (i / a, pf, suffix)
                    p -= 1
                a *= 1024
        return "??? " + suffix

    def dmrg(self, bond_dims, noises, dav_thrds, occs=None, bias=1.0, n_steps=30, conv_tol=1E-7, cutoff=1E-14):
        """Ground-State DMRG.

        Args:
            bond_dims : list(int), MPS bond dims for each sweep
            noises : list(double), noise for each sweep
            dav_thrds : list(double), Davidson convergence for each sweep
            occs : list(double) or None
                if occs = None, use FCI init MPS
                if occs = list(double), use occ init MPS
            bias : double, effective when occs is not None
                bias = 0.0: HF occ (e.g. occs => 2 2 2 ... 0 0 0)
                bias = 1.0: no bias (e.g. occs => 1.7 1.8 ... 0.1 0.05)
                bias = +inf: fully random (e.g. occs => 1 1 1 ... 1 1 1)
                
                increase bias if you want an init MPS with larger bond dim

                0 <= occ <= 2
                biased occ = 1 + (occ - 1) ** bias    (if occ > 1)
                           = 1 - (1 - occ) ** bias    (if occ < 1)
            n_steps : int, maximal number of sweeps
            conv_tol : double, energy convergence
            cutoff : cutoff of density matrix eigenvalues (default is the same as StackBlock)
        """

        if self.verbose >= 2:
            print('>>> START GS-DMRG <<<')
        t = time.perf_counter()

        # MPSInfo
        mps_info = MPSInfo(self.n_sites, self.hamil.vacuum,
                           self.target, self.hamil.basis)
        mps_info.tag = 'KET'
        if occs is None:
            if self.verbose >= 2:
                print("Using FCI INIT MPS")
            mps_info.set_bond_dimension(bond_dims[0])
        else:
            if self.verbose >= 2:
                print("Using occupation number INIT MPS")
            mps_info.set_bond_dimension_using_occ(
                bond_dims[0], VectorDouble(occs), bias=bias)
        mps = MPS(self.n_sites, 0, 2)
        mps.initialize(mps_info)
        mps.random_canonicalize()

        mps.save_mutable()
        mps.deallocate()
        mps_info.save_mutable()
        mps_info.deallocate_mutable()

        # MPO
        tx = time.perf_counter()
        mpo = MPOQC(self.hamil, QCTypes.Conventional)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True)
        self.mpo_orig = mpo

        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO
            mpo = ParallelMPO(mpo, self.prule)

        if self.verbose >= 3:
            print('MPO time = ', time.perf_counter() - tx)

        if self.print_statistics:
            print('GS MPO BOND DIMS = ', ''.join(
                ["%6d" % (x.m * x.n) for x in mpo.left_operator_names]))
            max_d = max(bond_dims)
            mps_info2 = MPSInfo(self.n_sites, self.hamil.vacuum,
                                self.target, self.hamil.basis)
            mps_info2.set_bond_dimension(max_d)
            _, mem2, disk = mpo.estimate_storage(mps_info2, 2)
            print("GS INIT MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info.left_dims]))
            print("GS EST MAX MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info2.left_dims]))
            print("GS EST PEAK MEM = ", GSDMRG.fmt_size(
                mem2), " SCRATCH = ", GSDMRG.fmt_size(disk))
            mps_info2.deallocate_mutable()
            mps_info2.deallocate()

        # DMRG
        me = MovingEnvironment(mpo, mps, mps, "DMRG")
        if self.dctr:
            me.delayed_contraction = OpNamesSet.normal_ops()
        tx = time.perf_counter()
        me.init_environments(self.verbose >= 4)
        if self.verbose >= 3:
            print('DMRG INIT time = ', time.perf_counter() - tx)
        dmrg = DMRG(me, VectorUBond(bond_dims), VectorDouble(noises))
        dmrg.davidson_conv_thrds = VectorDouble(dav_thrds)
        dmrg.noise_type = NoiseTypes.ReducedPerturbative
        dmrg.decomp_type = DecompositionTypes.DensityMatrix
        dmrg.iprint = max(self.verbose - 1, 0)
        dmrg.cutoff = cutoff
        dmrg.solve(n_steps, mps.center == 0, conv_tol)

        self.gs_energy = dmrg.energies[-1][0]
        self.bond_dim = bond_dims[-1]

        mps.save_data()
        mps_info.save_data(self.scratch + "/GS_MPS_INFO")
        mps_info.deallocate()

        if self.print_statistics:
            dmain, dseco, imain, iseco = Global.frame.peak_used_memory
            print("GS PEAK MEM USAGE:",
                  "DMEM = ", GSDMRG.fmt_size(dmain + dseco),
                  "(%.0f%%)" % (dmain * 100 / (dmain + dseco)),
                  "IMEM = ", GSDMRG.fmt_size(imain + iseco),
                  "(%.0f%%)" % (imain * 100 / (imain + iseco)))

        if self.verbose >= 1:
            print("=== GS Energy = %20.15f" % self.gs_energy)

        if self.verbose >= 2:
            print('>>> COMPLETE GS-DMRG | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        return self.gs_energy

    def expectation(self):
        """Expectation on ground-state MPS (after GS-DMRG)."""
        if self.verbose >= 2:
            print('>>> START expectation <<<')
        t = time.perf_counter()

        if MPI is not None:
            MPI.barrier()

        mps_info = MPSInfo(0)
        mps_info.load_data(self.scratch + "/GS_MPS_INFO")
        mps = MPS(mps_info)
        mps.load_data()

        # MPO
        tx = time.perf_counter()
        mpo = MPOQC(self.hamil, QCTypes.Conventional)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True)

        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO
            mpo = ParallelMPO(mpo, self.prule)

        if self.verbose >= 3:
            print('MPO time = ', time.perf_counter() - tx)

        if self.print_statistics:
            print('GS EXPECT MPO BOND DIMS = ', ''.join(
                ["%6d" % (x.m * x.n) for x in mpo.left_operator_names]))
            mps.info.load_mutable()
            _, mem2, disk = mpo.estimate_storage(mps_info, 2)
            print("GS MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps.info.left_dims]))
            print("GS EST PEAK MEM = ", GSDMRG.fmt_size(
                mem2), " SCRATCH = ", GSDMRG.fmt_size(disk))
            mps.info.deallocate_mutable()

        if self.verbose >= 1:
            print("=== GS Energy = %20.15f" % self.gs_energy)

        if self.verbose >= 2:
            print('>>> COMPLETE GS-DMRG | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        pme = MovingEnvironment(mpo, mps, mps, "EX")
        pme.init_environments(False)
        expect = Expect(pme, mps.info.bond_dim + 100, mps.info.bond_dim + 100)
        expect.iprint = max(self.verbose - 1, 0)
        ener = expect.solve(False, mps.center == 0)

        mps.save_data()
        mps_info.deallocate()
        mpo.deallocate()

        return ener

    # one-particle density matrix
    # return value:
    #     pdm[0, :, :] -> <AD_{i,alpha} A_{j,alpha}>
    #     pdm[1, :, :] -> < AD_{i,beta}  A_{j,beta}>
    def get_one_pdm(self, ridx=None):
        """1PDM on ground-state MPS (after GS-DMRG).
        ridx : reordering of orbitals (if no reordering, use ridx = None)"""
        if self.verbose >= 2:
            print('>>> START one-pdm <<<')
        t = time.perf_counter()

        if MPI is not None:
            MPI.barrier()

        mps_info = MPSInfo(0)
        mps_info.load_data(self.scratch + "/GS_MPS_INFO")
        mps = MPS(mps_info)
        mps.load_data()

        mps.info.load_mutable()
        max_bdim = max([x.n_states_total for x in mps.info.left_dims])
        if mps.info.bond_dim < max_bdim:
            mps.info.bond_dim = max_bdim
        max_bdim = max([x.n_states_total for x in mps.info.right_dims])
        if mps.info.bond_dim < max_bdim:
            mps.info.bond_dim = max_bdim

        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO

        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        if self.mpi is not None:
            pmpo = ParallelMPO(pmpo, self.pdmrule)

        # 1PDM
        pme = MovingEnvironment(pmpo, mps, mps, "1PDM")
        pme.init_environments(False)
        expect = Expect(pme, mps.info.bond_dim, mps.info.bond_dim)
        expect.iprint = max(self.verbose - 1, 0)
        expect.solve(True, mps.center == 0)
        if SpinLabel == SU2:
            dmr = expect.get_1pdm_spatial(self.n_sites)
            dm = np.array(dmr).copy()
        else:
            dmr = expect.get_1pdm(self.n_sites)
            dm = np.array(dmr).copy()
            dm = dm.reshape((self.n_sites, 2,
                             self.n_sites, 2))
            dm = np.transpose(dm, (0, 2, 1, 3))

        if ridx is not None:
            dm[:, :] = dm[ridx, :][:, ridx]

        mps.save_data()
        mps_info.deallocate()
        dmr.deallocate()
        pmpo.deallocate()

        if self.verbose >= 2:
            print('>>> COMPLETE one-pdm | Time = %.2f <<<' %
                  (time.perf_counter() - t))

        if SpinLabel == SU2:
            return np.concatenate([dm[None, :, :], dm[None, :, :]], axis=0) / 2
        else:
            return np.concatenate([dm[None, :, :, 0, 0], dm[None, :, :, 1, 1]], axis=0)

    def __del__(self):
        if self.hamil is not None:
            self.hamil.deallocate()
        if self.fcidump is not None:
            self.fcidump.deallocate()
        if self.mpo_orig is not None:
            self.mpo_orig.deallocate()
        release_memory()


if __name__ == "__main__":

    # parameters
    n_threads = 4
    hf_type = "RHF"  # RHF or UHF
    mpg = 'd2h'  # point group: d2h or c1
    scratch = './tmp'

    memory = 1E9  # in bytes
    verbose = 3
    do_ccsd = True # True: use ccsd init MPS; False: use random init MPS
    bias = 10

    # if n_sweeps is larger than len(bond_dims), the last value will be repeated
    bond_dims = [250, 250, 250, 500, 500, 1000]
    # unit : norm(wfn) ** 2 (same as StackBlock)
    noises = [1E-6, 1E-7, 1E-8, 1E-9, 1E-9]
    # unit : norm(wfn) ** 2 (same as StackBlock)
    dav_thrds = [1E-6, 1E-6, 1E-7, 1E-7, 1E-8, 1E-8]
    n_sweeps = 30
    conv_tol = 1E-8

    import os
    if MPI is None or MPI.rank == 0:
        if not os.path.isdir(scratch):
            os.makedirs(scratch)
    if MPI is not None:
        MPI.barrier()
    os.environ['TMPDIR'] = scratch

    from pyscf import gto, scf, symm, ao2mo, cc

    # H chain
    N = 10
    BOHR = 0.52917721092  # Angstroms
    R = 1.8 * BOHR
    mol = gto.M(atom=[['H', (0, 0, i * R)] for i in range(N)],
                basis='sto6g', verbose=0, symmetry=mpg)
    pg = mol.symmetry.lower()

    if hf_type == "RHF":
        mf = scf.RHF(mol)
    elif hf_type == "UHF":
        assert SpinLabel == SZ
        mf = scf.UHF(mol)

    if MPI is None or MPI.rank == 0:
        ener = mf.kernel()
    else:
        ener = 0
        mf.mo_coeff = None

    if MPI is not None:
        ener = comm.bcast(ener, root=0)
        mf.mo_coeff = comm.bcast(mf.mo_coeff, root=0)

    _print("SCF Energy = %20.15f" % ener)
    _print(("NON-" if SpinLabel == SZ else "") + "SPIN-ADAPTED")

    pg = mol.symmetry.lower()
    if pg == 'd2h':
        fcidump_sym = ["Ag", "B3u", "B2u", "B1g", "B1u", "B2g", "B3g", "Au"]
    elif pg == 'c1':
        fcidump_sym = ["A"]
    else:
        raise GSDMRGError("Point group %d not supported yet!" % pg)

    mo_coeff = mf.mo_coeff

    if hf_type == "RHF":

        n_mo = mo_coeff.shape[1]

        orb_sym_str = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff)
        orb_sym = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str])

        h1e = mo_coeff.T @ mf.get_hcore() @ mo_coeff
        g2e = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff), n_mo)
        ecore = mol.energy_nuc()
        na = nb = mol.nelectron // 2

    else:

        mo_coeff_a, mo_coeff_b = mo_coeff[0], mo_coeff[1]
        n_mo = mo_coeff_b.shape[1]

        orb_sym_str_a = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff_a)
        orb_sym_a = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str_a])

        orb_sym = orb_sym_a

        h1ea = mo_coeff_a.T @ mf.get_hcore() @ mo_coeff_a
        h1eb = mo_coeff_b.T @ mf.get_hcore() @ mo_coeff_b
        g2eaa = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_a), n_mo)
        g2ebb = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff_b), n_mo)
        g2eab = ao2mo.kernel(
            mol, [mo_coeff_a, mo_coeff_a, mo_coeff_b, mo_coeff_b])
        h1e = (h1ea, h1eb)
        g2e = (g2eaa, g2ebb, g2eab)
        ecore = mol.energy_nuc()
        na, nb = mol.nelec
    
    if do_ccsd:
        if MPI is None or MPI.rank == 0:
            mcc = cc.CCSD(mf)
            mcc.kernel()
            dmmo = mcc.make_rdm1()
            if hf_type == "RHF":
                occs = np.diag(dmmo)
            else:
                occs = np.diag(dmmo[0]) + np.diag(dmmo[1])
        else:
            occs = None
        if MPI is not None:
            occs = comm.bcast(occs, root=0)
        _print('OCCS = ', occs)
    else:
        occs = None

    dmrg = GSDMRG(scratch=scratch, memory=memory,
                  verbose=verbose, omp_threads=n_threads, mpi=MPI)
    dmrg.init_hamiltonian(pg, n_sites=n_mo, n_elec=na + nb, twos=na - nb, isym=1,
                          orb_sym=orb_sym, e_core=ecore, h1e=h1e, g2e=g2e)
    # calculate ground-state energy
    ener = dmrg.dmrg(bond_dims=bond_dims, noises=noises, dav_thrds=dav_thrds, occs=occs, bias=bias,
                     n_steps=n_sweeps, conv_tol=conv_tol)
    # calculate pdm before changing FCIDUMP
    pdm = dmrg.get_one_pdm()

    # change integrals
    dmrg.init_hamiltonian(pg, n_sites=n_mo, n_elec=na + nb, twos=na - nb, isym=1,
                          orb_sym=orb_sym, e_core=ecore - 5, h1e=h1e, g2e=g2e)
    # calculate expectation of new FCIDUMP on ground-state MPS
    expect = dmrg.expectation()

    _print('GS ENERGY = %20.15f' % ener)
    _print('EXPECTATION = %20.15f' % expect)
    _print('pdm =', pdm)
    assert abs(ener - (-5.424385376325292)) < conv_tol
    assert abs(expect - (-5.424385376325292 - 5)) < conv_tol
    assert abs(pdm[0, 0, 0] - 9.90728423e-01) < np.sqrt(conv_tol)

    del dmrg  # IMPORTANT!!! --> to release stack memory




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
Initialization of Hamiltonian, MPS, and MPO.
"""

import contextlib
import numpy as np
import shutil

from .io import MPSTools, MPOTools
from .core import MPS as PYMPS, Tensor, SubTensor

from block2 import SZ, Global
from block2 import init_memory, release_memory, set_mkl_num_threads
from block2 import VectorUInt8, VectorUBond, VectorDouble, PointGroup
from block2 import Random, FCIDUMP, QCTypes, SeqTypes, NoiseTypes
from block2.sz import HamiltonianQC, MPS, MPSInfo, AncillaMPSInfo
from block2.sz import PDM1MPOQC, AncillaMPO, SimplifiedMPO, Rule, RuleQC, MPOQC
from block2.sz import DMRG, MovingEnvironment, NoTransposeRule

class HamilTools:
    """
    Initialization of different type of systems.
    """
    def __init__(self, hamil):
        self.hamil = hamil

    @staticmethod
    @contextlib.contextmanager
    def _init(scratch='./my_tmp', rand_seed=1234, memory=int(1E9), n_threads=1):
        Random.rand_seed(rand_seed)
        init_memory(isize=int(memory * 0.1),
                    dsize=int(memory * 0.9), save_dir=scratch)
        import os
        empty_scratch = len(os.listdir(scratch)) == 0
        set_mkl_num_threads(n_threads)

        yield ()

        release_memory()
        if empty_scratch:
            shutil.rmtree(scratch)

    @staticmethod
    @contextlib.contextmanager
    def _from_fcidump(fcidump, pg='d2h'):
        swap_pg = getattr(PointGroup, "swap_" + pg)
        vacuum = SZ(0, 0, 0)
        target = SZ(fcidump.n_elec, fcidump.twos, swap_pg(fcidump.isym))

        n_sites = fcidump.n_sites
        orb_sym = VectorUInt8(map(swap_pg, fcidump.orb_sym))
        hamil = HamiltonianQC(vacuum, n_sites, orb_sym, fcidump)

        yield HamilTools(hamil)

        hamil.deallocate()
        fcidump.deallocate()
    
    @staticmethod
    @contextlib.contextmanager
    def from_fcidump(filename, pg='d2h', **kwargs):
        """
        Read quantum chemistry system from FCIDUMP file.
        """
        with HamilTools._init(**kwargs) as ():
            fcidump = FCIDUMP()
            fcidump.read(filename)
            with HamilTools._from_fcidump(fcidump, pg=pg) as hamil:
                yield hamil
    
    @staticmethod
    @contextlib.contextmanager
    def hchain(n_sites, r=1.8, pg_reorder=True, **kwargs):
        """
        1D Hydrogen chain model. r in bohr.
        """
        if 'scratch' in kwargs:
            import os
            if not os.path.isdir(kwargs['scratch']):
                os.mkdir(kwargs['scratch'])
            os.environ['TMPDIR'] = kwargs['scratch']

        from pyscf import gto, scf, symm, ao2mo

        BOHR = 0.52917721092  # Angstroms
        r = r * BOHR
        mol = gto.M(atom=[['H', (i * r, 0, 0)] for i in range(n_sites)],
            basis='sto6g', verbose=0, symmetry='d2h')
        pg = mol.symmetry.lower()

        # Reorder
        if pg == 'd2h':
            fcidump_sym = ["Ag", "B3u", "B2u", "B1g", "B1u", "B2g", "B3g", "Au"]
            optimal_reorder = ["Ag", "B1u", "B3u",
                            "B2g", "B2u", "B3g", "B1g", "Au"]
        elif pg == 'c1':
            fcidump_sym = ["A"]
            optimal_reorder = ["A"]
        else:
            assert False
        
        # SCF
        m = scf.RHF(mol)
        m.kernel()
        mo_coeff = m.mo_coeff
        n_ao = mo_coeff.shape[0]
        n_mo = mo_coeff.shape[1]

        orb_sym_str = symm.label_orb_symm(
            mol, mol.irrep_name, mol.symm_orb, mo_coeff)
        orb_sym = np.array([fcidump_sym.index(i) + 1 for i in orb_sym_str])

        # Sort the orbitals by symmetry for more efficient DMRG
        if pg_reorder:
            idx = np.argsort([optimal_reorder.index(i) for i in orb_sym_str])
            orb_sym = orb_sym[idx]
            mo_coeff = mo_coeff[:, idx]
            ridx = np.argsort(idx)
        else:
            ridx = np.array(list(range(n_mo)), dtype=int)

        h1e = mo_coeff.T @ m.get_hcore() @ mo_coeff
        g2e = ao2mo.restore(8, ao2mo.kernel(mol, mo_coeff), n_mo)
        ecore = mol.energy_nuc()
        del m, mol

        with HamilTools._init(**kwargs) as ():
            fcidump = FCIDUMP()
            n_elec = n_sites
            tol = 1E-13
            mh1e = np.zeros((n_sites * (n_sites + 1) // 2))
            k = 0
            for i in range(0, n_sites):
                for j in range(0, i + 1):
                    assert abs(h1e[i, j] - h1e[j, i]) < tol
                    mh1e[k] = h1e[i, j]
                    k += 1
            mg2e = g2e.flatten().copy()
            mh1e[np.abs(mh1e) < tol] = 0.0
            mg2e[np.abs(mg2e) < tol] = 0.0
            fcidump.initialize_su2(n_sites, n_elec, 0, 1, ecore, mh1e, mg2e)
            fcidump.orb_sym = VectorUInt8(orb_sym)
            with HamilTools._from_fcidump(fcidump, pg=pg) as hamil:
                yield hamil

    @staticmethod
    @contextlib.contextmanager
    def hubbard(n_sites, u=2, t=1, **kwargs):
        """
        1D Hubbard model.
        """
        with HamilTools._init(**kwargs) as ():
            fcidump = FCIDUMP()
            h1e = np.zeros((n_sites * (n_sites + 1) // 2), dtype=float)
            g2e = np.zeros((n_sites * (n_sites + 1) // 2 * (n_sites * (n_sites + 1) // 2 + 1) // 2), dtype=float)
            ij, kl, ijkl = 0, 0, 0
            for i in range(0, n_sites):
                for j in range(0, i + 1):
                    if abs(i - j) == 1:
                        h1e[ij] = t
                    kl = 0
                    for k in range(0, n_sites):
                        for l in range(0, k + 1):
                            if ij >= kl:
                                if i == j and k == l and i == k:
                                    g2e[ijkl] = u
                                ijkl += 1
                            kl += 1
                    ij += 1

            fcidump.initialize_su2(n_sites, n_sites, 0, 1, 0, h1e, g2e)
            fcidump.orb_sym = VectorUInt8([1] * n_sites)
            with HamilTools._from_fcidump(fcidump, pg='c1') as hamil:
                yield hamil

    def get_mpo(self, mode="NC", mu=0.0, ancilla=False):
        self.hamil.mu = mu
        bmpo = MPOQC(self.hamil, QCTypes.NC if mode == "NC" else QCTypes.CN)
        self.hamil.mu = 0.0
        if ancilla:
            bmpo = AncillaMPO(bmpo)
        mpo = MPOTools.from_block2(bmpo)
        bmpo.deallocate()
        return mpo
    
    @contextlib.contextmanager
    def get_thermal_limit_mps_block2(self, dot=2):
        hamil = self.hamil
        vacuum = hamil.vacuum
        target = SZ(hamil.n_sites * 2, hamil.fcidump.twos, 0)

        mps_info = AncillaMPSInfo(hamil.n_sites, vacuum, target, hamil.basis)
        mps_info.set_thermal_limit()
        mps = MPS(hamil.n_sites * 2, 0, dot)
        mps.initialize(mps_info)
        mps.fill_thermal_limit()
        mps.canonicalize()

        mps.save_mutable()
        mps.deallocate()
        mps_info.save_mutable()
        mps_info.deallocate_mutable()

        yield mps
        mps_info.deallocate()
    
    @contextlib.contextmanager
    def get_init_mps_block2(self, bond_dim=250, dot=2):

        hamil = self.hamil
        vacuum = hamil.vacuum
        target = SZ(hamil.fcidump.n_elec, hamil.fcidump.twos, 0)

        mps_info = MPSInfo(hamil.n_sites, vacuum, target, hamil.basis)
        mps_info.set_bond_dimension(bond_dim)
        mps = MPS(hamil.n_sites, 0, dot)
        mps.initialize(mps_info)
        mps.random_canonicalize()

        mps.save_mutable()
        mps.deallocate()
        mps_info.save_mutable()
        mps_info.deallocate_mutable()

        yield mps
        mps_info.deallocate()

    def get_init_mps(self, bond_dim=250):
        with self.get_init_mps_block2(bond_dim=bond_dim, dot=1) as mps:
            xmps = MPSTools.from_block2(mps)
        return xmps
    
    def get_thermal_limit_mps(self):
        with self.get_thermal_limit_mps_block2(dot=1) as mps:
            xmps = MPSTools.from_block2(mps)
        return xmps

    @staticmethod
    def get_determinant_mps(det, orb_sym=None):
        """
        MPS from SZ determinant.
        det = [0a, 0b, 1a, 1b, ...] where ia/ib = 0 (empty) or 1 (occupied).
        """
        n_sites = len(det) // 2
        tensors = []
        pq = None
        for i in range(n_sites):
            if det[i * 2] == 1 and det[i * 2 + 1] == 1:
                q = SZ(2, 0, 0)
            elif det[i * 2] == 1 and det[i * 2 + 1] == 0:
                q = SZ(1, 1, 0 if orb_sym is None else orb_sym[i])
            elif det[i * 2] == 0 and det[i * 2 + 1] == 1:
                q = SZ(1, -1, 0 if orb_sym is None else orb_sym[i])
            else:
                q = SZ(0, 0, 0)
            if i == 0:
                ql = (q, q)
                rd = np.array([[1.0]])
            elif i == n_sites - 1:
                ql = (pq, q)
                rd = np.array([[1.0]])
            else:
                ql = (pq, q, pq + q)
                rd = np.array([[[1.0]]])
            pq = pq + q if pq is not None else q
            ts = Tensor(blocks=[SubTensor(q_labels=ql, reduced=rd)])
            tensors.append(ts)
        return PYMPS(tensors=tensors)

    @contextlib.contextmanager
    def get_ground_state_mps_block2(self, bond_dim=250, n_sweeps=20):
        hamil = self.hamil
        with self.get_init_mps_block2(bond_dim=bond_dim, dot=2) as mps:
            mpo = MPOQC(hamil, QCTypes.Conventional)
            mpo = SimplifiedMPO(mpo, RuleQC(), True)

            me = MovingEnvironment(mpo, mps, mps, "DMRG")
            me.init_environments(True)
            dmrg = DMRG(me, VectorUBond([bond_dim]), VectorDouble([1E-6, 0]))
            energy = dmrg.solve(n_sweeps, mps.center == 0)

            print("Ground State Energy = ", energy)
            yield mps
            mpo.deallocate()

    def get_ground_state_mps(self, bond_dim=250, n_sweeps=20):
        with self.get_ground_state_mps_block2(bond_dim=bond_dim, n_sweeps=n_sweeps) as mps:
            if mps.center == 0:
                mps.dot = 1
            elif mps.center == mps.n_sites - 2:
                mps.center = mps.n_sites - 1
                mps.dot = 1
            xmps = MPSTools.from_block2(mps)
        return xmps



#! /usr/bin/env python
"""
Read StackBlock MPS,
Write block2 MPS.

Needs pyblock (https://github.com/hczhai/pyblock).
Check README of block2 if you get `double free` error from python
when import pyblock and block2 in the same script.

Author:
    Huanchen Zhai
    Apr 15, 2021
"""

try:
    from pyblock2.driver.parser import parse
except ImportError:
    from parser import parse

import sys
import os
import numpy as np

if len(sys.argv) > 1:
    arg_dic = {}
    for i in range(1, len(sys.argv)):
        if sys.argv[i] in ['-su2', '-sz', '-expect', '-reduntant']:
            arg_dic[sys.argv[i][1:]] = ''
        elif sys.argv[i].startswith('-'):
            arg_dic[sys.argv[i][1:]] = sys.argv[i + 1]
        elif i == 1:
            arg_dic['config'] = sys.argv[i]
else:
    raise ValueError("""
        Usage:
            (A) python readwfn.py -config dmrg.conf -out ./out
            (B) python readwfn.py dmrg.conf
            (C) python readwfn.py dmrg.conf -expect
            (D) python readwfn.py dmrg.conf -reduntant
            (E) python readwfn.py -integral FCIDUMP -prefix ./scratch -dot 2 -su2
            (F) python readwfn.py -integral FCIDUMP -prefix ./scratch -dot 2 -sz
            (G) python readwfn.py ... -sym c1

        Args:
            config: StackBlock input file
            out: dir for storing block2 MPS
            expect: if given, the energy expectation value of MPS
                is calculated using block2 and printed at the end
            reduntant: if given, the reduntant parameter in the
                StackBlock MPS will be retained.
                Note that removing reduntant parameters do not
                affect quality of MPS.

            when no config file is given/available:
                integral: path to integral file
                prefix: path to StackBlock MPS (does not include 'node0')
                dot: 1 or 2 dot MPS
                su2/sz: spin-adapted or non-spin-adapted
                sym: point group name (default is d2h)
                    note that d2h also works for c1

            when the prefix is set in config file using relative path,
                it will be considered as relative to the dir of dmrg.conf
                but not relative to the current dir

        If StackBlock is run with reordered integrals, before running this script,
        a reordered FCIDUMP should be generated using
            python gaopt.py -ord RestartReorder.dat -wint FCIDUMP.NEW
        And then use python readwfn.py -integral FCIDUMP.NEW ...
    """)

scratch = './node0'
integral = 'FCIDUMP'
dot = 1
su2 = True
sym = "d2h"
nelec = None
spin = None
hf_occ = None
out_dir = "./out"
expect = "expect" in arg_dic
redunt = "reduntant" in arg_dic
mps_tags = ["KET"]
if "config" in arg_dic:
    config = arg_dic["config"]
    dic = parse(config)
    dd = os.path.dirname(config)
    scratch = dic.get("prefix", "./") + "/node0"
    if not os.path.isabs(scratch):
        scratch = ('.' if dd == '' else dd) + "/" + scratch
    integral = dic["orbitals"]
    if not os.path.isabs(integral):
        integral = ('.' if dd == '' else dd) + "/" + integral
    dot = 1 if "twodot_to_onedot" in dic or "onedot" in dic else 2
    su2 = "nonspinadapted" not in dic
    mps_tags = dic.get("mps_tags", "KET").split()
    sym = dic.get("sym", "d2h")
    if "spin" in dic:
        spin = int(dic["spin"])
    if "nelec" in dic:
        nelec = int(dic["nelec"])
    if "hf_occ" in dic:
        hf_occ = dic["hf_occ"]
if "prefix" in arg_dic:
    scratch = arg_dic["prefix"] + "/node0"
if "integral" in arg_dic:
    integral = arg_dic["integral"]
if "out" in arg_dic:
    out_dir = arg_dic["out"]
if os.path.isfile(scratch + "/wave-0-3.0.0.tmp"):
    dot = 1 if open(scratch + "/wave-0-3.0.0.tmp", 'rb').read(41)[-1] else 2
elif os.path.isfile(scratch + "/wave-0-1.0.0.tmp"):
    dot = 1 if open(scratch + "/wave-0-1.0.0.tmp", 'rb').read(41)[-1] else 2
if "su2" in arg_dic:
    su2 = True
if "sz" in arg_dic:
    su2 = False
if "dot" in arg_dic:
    dot = int(arg_dic["dot"])
if "sym" in arg_dic:
    sym = arg_dic["sym"]

from pyblock.qchem import DMRGDataPage, BlockHamiltonian
from pyblock.algorithm import DMRG
from pyblock.qchem.core import BlockSymmetry

from block import VectorInt, VectorMatrix, load_rotation_matrix
from block.dmrg import SweepParams
from block.io import Global
from block.symmetry import StateInfo, state_tensor_product_target, state_tensor_product
from block.operator import Wavefunction

page = DMRGDataPage(save_dir=scratch)

opts = dict(fcidump=integral, pg=sym,
            su2=su2, output_level=-1, memory=25000,
            omp_threads=1, mkl_threads=1, page=page)

if spin is not None:
    opts['spin'] = spin
if nelec is not None:
    opts['nelec'] = nelec
if hf_occ is not None:
    opts['hf_occ'] = hf_occ

hamil_cxt = BlockHamiltonian.get(**opts)
hamil = hamil_cxt.__enter__()

site_sts = hamil.site_state_info
swpp = SweepParams()
Global.dmrginp.load_prefix = scratch
Global.dmrginp.save_prefix = scratch
forward = True
center = hamil.n_sites - dot - 1

stls = []
for i in range(0, center):
    st = StateInfo()
    st.load(forward, VectorInt(range(i + 1)), -1)
    stls.append(st)
strs = []
for i in range(center + dot, hamil.n_sites):
    st = StateInfo()
    st.load(not forward, VectorInt(range(i, hamil.n_sites)), -1)
    strs.append(st)

if su2:
    wave_sites = VectorInt(range(0, center + 1))
else:
    wave_sites = VectorInt(range(0, 2 * (center + 1)))
wll = state_tensor_product(stls[center - 1], hamil.site_state_info[center][0])
wll.collect_quanta()
if dot == 1:
    wrr = strs[0]
else:
    wrr = state_tensor_product(strs[0], hamil.site_state_info[center + 1][0])
    wrr.collect_quanta()
llrr = state_tensor_product_target(wll, wrr)
wave = Wavefunction()
target_state_info = BlockSymmetry.to_state_info([(hamil.target, 1)])
wave.initialize(target_state_info.quanta, wll, wrr, dot == 1)
wave.load_wavefunction_info(llrr, wave_sites, 0, False)
assert wave.onedot == (dot == 1)

rots = []
for i in range(1, center):
    rotation_matrix = VectorMatrix()
    if su2:
        load_rotation_matrix(VectorInt(range(i + 1)), rotation_matrix, -1)
    else:
        load_rotation_matrix(
            VectorInt(range(2 * (i + 1))), rotation_matrix, -1)
    rots.append(rotation_matrix)

from block2 import SU2, SZ, Global, VectorDouble, VectorUInt16
from block2 import init_memory, set_mkl_num_threads, QCTypes
from block2 import PointGroup, VectorUInt8, FCIDUMP, SeqTypes, NoiseTypes

if su2:
    SX = SU2
    from block2.su2 import MPSInfo, MPS, StateInfo, HamiltonianQC, Expect, SparseMatrix
    from block2.su2 import MPOQC, RuleQC, SimplifiedMPO, MovingEnvironment, DMRG, OperatorFunctions
else:
    SX = SZ
    from block2.sz import MPSInfo, MPS, StateInfo, HamiltonianQC, Expect, SparseMatrix
    from block2.sz import MPOQC, RuleQC, SimplifiedMPO, MovingEnvironment, DMRG, OperatorFunctions

if not os.path.isdir(out_dir):
    os.mkdir(out_dir)

swap_pg = getattr(PointGroup, "swap_" + sym)
npg = {'d2h': 8, 'c2h': 4, 'c2v': 4, 'd2': 4,
       'ci': 2, 'c2': 2, 'cs': 2, 'c1': 1}[sym]
inv_swap_pg = list(np.argsort([swap_pg(x + 1) for x in range(0, npg)]) + 1)

memory = int(10 * 1E9)
init_memory(isize=int(memory * 0.1), dsize=int(memory * 0.9), save_dir=out_dir)
fcidump = FCIDUMP()
fcidump.read(integral)
n_sites = fcidump.n_sites

twos = fcidump.twos if spin is None else spin
nelec = fcidump.n_elec if nelec is None else nelec
vacuum = SX(0)
# singlet embedding
left_vacuum = vacuum if twos == 0 else SX(twos, twos, 0)
target = SX(nelec + twos, 0, swap_pg(fcidump.isym))

orb_sym = VectorUInt8(map(swap_pg, fcidump.orb_sym))
hamil = HamiltonianQC(vacuum, n_sites, orb_sym, fcidump)
hamil.opf.seq.mode = SeqTypes.Simple
mps_info = MPSInfo(n_sites, vacuum, target, hamil.basis)
mps_info.tag = mps_tags[0]
if redunt:
    mps_info.set_bond_dimension_full_fci(left_vacuum, vacuum)
else:
    mps_info.set_bond_dimension_fci(left_vacuum, vacuum)

mps_info.left_dims[0] = StateInfo(left_vacuum)
mps_info.save_left_dims(0)
for i in range(0, center + 1):
    st = StateInfo()
    xst = stls[i] if i != center else wll
    n = len(xst.quanta)
    st.allocate(n)
    st.n = n
    for j in range(n):
        st.quanta[j] = SX(xst.quanta[j].n, xst.quanta[j].s.irrep,
                          swap_pg(xst.quanta[j].symm.irrep + 1))
        st.n_states[j] = xst.n_states[j]
        if not redunt and mps_info.left_dims_fci[i + 1].find_state(st.quanta[j]) == -1:
            st.n_states[j] = 0
    st.sort_states()
    if not redunt:
        st.collect()
    mps_info.left_dims[i + 1] = st
    mps_info.save_left_dims(i + 1)

mps_info.right_dims[n_sites] = StateInfo(vacuum)
mps_info.save_right_dims(n_sites)
for i in range(center + 1, n_sites):
    st = StateInfo()
    xst = strs[i - center - dot] if i != center + dot - 1 else wrr
    n = len(xst.quanta)
    st.allocate(n)
    st.n = n
    for j in range(n):
        st.quanta[j] = SX(xst.quanta[j].n, xst.quanta[j].s.irrep,
                          swap_pg(xst.quanta[j].symm.irrep + 1))
        st.n_states[j] = xst.n_states[j]
    st.sort_states()
    mps_info.right_dims[i] = st
    mps_info.save_right_dims(i)

mps_info.save_mutable()

mps = MPS(n_sites, center, dot)
mps.initialize(mps_info)


def swap_order_left(idx):
    dd = {}
    mps_info.load_left_dims(idx)
    mps_info.load_left_dims(idx + 1)
    l, m, r = mps_info.left_dims[idx], hamil.basis[idx], mps_info.left_dims[idx + 1]
    clm = StateInfo.get_connection_info(l, m, r)
    for ik in range(r.n):
        bbed = clm.n if ik == r.n - 1 else clm.n_states[ik + 1]
        dx = []
        g = 0
        for bb in range(clm.n_states[ik], bbed):
            ibba = clm.quanta[bb].data >> 16
            ibbb = clm.quanta[bb].data & 0xFFFF
            nx = l.n_states[ibba] * m.n_states[ibbb]
            dx.append((l.quanta[ibba], m.quanta[ibbb], g, g + nx))
            g += nx
        dx.sort(key=lambda x: (
            x[0].n, x[0].twos, inv_swap_pg[x[0].pg], x[1].n, x[1].twos, inv_swap_pg[x[1].pg]))
        pp = []
        for xx in dx:
            pp += list(range(xx[2], xx[3]))
        dd[r.quanta[ik]] = np.argsort(pp)
    return dd


def swap_order_right(idx):
    dd = {}
    mps_info.load_right_dims(idx)
    mps_info.load_right_dims(idx + 1)
    l, m, r = mps_info.right_dims[idx], hamil.basis[idx], mps_info.right_dims[idx + 1]
    clm = StateInfo.get_connection_info(m, r, l)
    for ik in range(l.n):
        bbed = clm.n if ik == l.n - 1 else clm.n_states[ik + 1]
        dx = []
        g = 0
        for bb in range(clm.n_states[ik], bbed):
            ibba = clm.quanta[bb].data >> 16
            ibbb = clm.quanta[bb].data & 0xFFFF
            nx = m.n_states[ibba] * r.n_states[ibbb]
            dx.append((m.quanta[ibba], r.quanta[ibbb], g, g + nx))
            g += nx
        dx.sort(key=lambda x: (
            x[1].n, x[1].twos, inv_swap_pg[x[1].pg], x[0].n, x[0].twos, inv_swap_pg[x[0].pg]))
        pp = []
        for xx in dx:
            pp += list(range(xx[2], xx[3]))
        dd[l.quanta[ik]] = np.argsort(pp)
    return dd


if su2:
    if twos == 0:
        mps.tensors[0].data = np.array([1.0, 1.0, 1.0])
    else:
        # singlet embedding
        assert mps.tensors[0].info.n == len(mps.tensors[0].data)
        mat = np.zeros_like(np.array(mps.tensors[0].data))
        for ix in range(mps.tensors[0].info.n):
            q = mps.tensors[0].info.quanta[ix]
            mat[ix] = -1 if (twos % 2 and q.twos == twos +
                             1) or (not twos % 2 and q.twos == twos - 1) else 1
        mps.tensors[0].data = mat
    mps.tensors[n_sites - 1].data = np.array([1.0, 1.0, 1.0])
else:
    mps.tensors[0].data = np.array([1.0, 1.0, 1.0, 1.0])
    mps.tensors[n_sites - 1].data = np.array([1.0, 1.0, 1.0, 1.0])

for i in range(1, center):
    stl = stls[i - 1]
    str = site_sts[i][0]
    st = state_tensor_product(stl, str)
    st.collect_quanta()
    assert len(st.quanta) == len(rots[i - 1])
    rot = rots[i - 1]
    mps.tensors[i].data[:] = 0
    swod = swap_order_left(i)
    for k in range(len(rot)):
        if rot[k].cols == 0:
            continue
        xq = st.quanta[k]
        q = SX(xq.n, xq.s.irrep, swap_pg(xq.symm.irrep + 1))
        iq = mps.tensors[i].info.find_state(q)
        if not redunt and not (iq != -1 and iq < mps.tensors[i].info.n):
            continue
        assert iq != -1 and iq < mps.tensors[i].info.n
        mat = mps.tensors[i][iq]
        assert rot[k].ref.size == np.array(mat).size
        assert mat.m == rot[k].ref.shape[0]
        assert mat.n == rot[k].ref.shape[1]
        xx = np.array(rot[k].ref, copy=True)
        mps.tensors[i][iq] = np.ascontiguousarray(xx[swod[q], :])

wfn = mps.tensors[center]
wfn.data[:] = 0

swl = swap_order_left(center)
swr = swap_order_right(center + 1)
for (il, ir), mat in wave.non_zero_blocks:
    xql = wll.quanta[il]
    xqr = wrr.quanta[ir]
    ql = SX(xql.n, xql.s.irrep, swap_pg(xql.symm.irrep + 1))
    qr = SX(xqr.n, xqr.s.irrep, swap_pg(xqr.symm.irrep + 1))
    q = wfn.info.delta_quantum.combine(ql, -qr)
    iq = wfn.info.find_state(q)
    if not redunt and iq == -1:
        continue
    assert iq != -1
    xmat = wfn[iq]
    assert xmat.m == mat.ref.shape[0]
    assert xmat.n == mat.ref.shape[1]
    xx = np.array(mat.ref, copy=True)
    xx = xx[swl[ql], :][:, swr[qr]]
    f = -1 if ql.twos == -2 or ql.twos == 2 else 1
    if (not su2) and dot == 2 and qr.n == 2 and qr.twos == 0 and xmat.n == 4:
        xx[:, 1:3] *= -1
    elif (not su2) and dot == 2 and qr.n == 2 and qr.twos == 0 and qr.pg == orb_sym[-1] ^ orb_sym[-2]:
        xx[:, :] *= -1
    wfn[iq] = f * np.ascontiguousarray(xx)

mps.save_mutable()
mps.save_data()
mps.deallocate()
max_bdim = max([x.n_states_total for x in mps_info.left_dims])
if mps_info.bond_dim < max_bdim:
    mps_info.bond_dim = max_bdim
max_bdim = max([x.n_states_total for x in mps_info.right_dims])
if mps_info.bond_dim < max_bdim:
    mps_info.bond_dim = max_bdim
mps_info.save_data(out_dir + '/mps_info.bin')
mps_info.save_data(out_dir + '/%s-mps_info.bin' % mps_tags[0])

if expect:
    mpo = MPOQC(hamil, QCTypes.Conventional)
    mpo = SimplifiedMPO(mpo, RuleQC(), True)
    me = MovingEnvironment(mpo, mps, mps, "DMRG")
    me.init_environments(False)
    ex = Expect(me, max_bdim, max_bdim).solve(False)
    print(ex)



#! /usr/bin/env python

"""
parser for converting stackblock input and libDMET integral to block2.

Author:
    Huanchen Zhai
    Zhi-Hao Cui
"""

from block2 import FCIDUMP
from block2 import VectorUInt8
import numpy as np

KNOWN_KEYS = {"nelec", "spin", "hf_occ", "schedule", "maxiter",
              "twodot_to_onedot", "twodot", "onedot", "zerodot", "sweep_tol",
              "orbitals", "warmup", "nroots", "outputlevel", "prefix",
              "noreorder", "fiedler", "reorder", "gaopt", "nofiedler", "irrep_reorder",
              "num_thrds", "mkl_thrds", "mem", "intmem", "oh", "nonspinadapted",
              "onepdm", "fullrestart", "restart_onepdm", "restart_oh",
              "twopdm", "restart_twopdm", "startM", "maxM", "symmetrize_ints",
              "diag_twopdm", "restart_diag_twopdm",
              "occ", "bias", "cbias", "correlation", "restart_correlation",
              "lowmem_noise", "dm_noise", "conn_centers", "restart_dir", "cutoff",
              "sym", "irrep", "weights", "statespecific", "mps_tags",
              "tran_onepdm", "tran_twopdm", "restart_tran_onepdm",
              "restart_tran_twopdm", "soc", "overlap", "tran_oh",
              "nat_orbs", "nat_km_reorder", "nat_positive_def",
              "restart_tran_oh", "restart_dir_per_sweep", "fp_cps_cutoff",
              "full_fci_space", "trans_mps_info", "trunc_type", "decomp_type",
              "orbital_rotation", "delta_t", "target_t", "te_type",
              "read_mps_tags", "compression", "random_mps_init", "trans_mps_to_sz",
              "trans_mps_to_singlet_embedding", "trans_mps_from_singlet_embedding",
              "copy_mps", "restart_copy_mps", "sample", "restart_sample", "resolve_twosz",
              "extrapolation", "cached_contraction", "singlet_embedding", "normalize_mps",
              "dmrgfci", "mrci", "mrcis", "mrcisd", "mrcisdt", "casci", "nevpt2", "nevpt2s",
              "nevpt2sd", "nevpt2-ijrs", "nevpt2-ij", "nevpt2-rs", "nevpt2-ijr",
              "nevpt2-rsi", "nevpt2-ir", "nevpt2-i", "nevpt2-r",
              "big_site", "stopt_dmrg", "stopt_compression", "stopt_sampling",
              "model", "k_symmetry", "k_irrep", "k_mod", "init_mps_center",
              "use_complex", "real_density_matrix", "expt_algo_type",
              "davidson_max_iter", "davidson_soft_max_iter", "one_body_parallel_rule",
              "n_sub_sweeps", "complex_mps", "split_states", "trans_mps_to_complex",
              "use_general_spin", "trans_integral_to_spin_orbital", "store_wfn_spectra"}

REORDER_KEYS = {"noreorder", "fiedler", "reorder", "gaopt", "nofiedler",
                "irrep_reorder"}

DYN_CORR_KEYS = {"dmrgfci", "mrci", "mrcis", "mrcisd", "mrcisdt", "casci", "nevpt2",
                 "nevpt2s", "nevpt2sd"}

NEVPT_KEYS = {"nevpt2", "nevpt2s", "nevpt2sd"}

RESTART_KEYS = {"restart_onepdm", "restart_twopdm", "restart_oh",
                "restart_correlation", "restart_tran_onepdm", "restart_tran_twopdm",
                "restart_tran_oh", "restart_tran_oh", "statespecific",
                "restart_copy_mps", "restart_sample", "orbital_rotation"}

GAOPT_KEYS = {"maxcomm", "maxgen", "maxcell",
              "cloning", "mutation", "elite", "scale", "method"}


def parse(fname):
    """
    parse a stackblock input file.

    Args:
        fname: stackblock input config file.

    Returns:
        dic: dictionary of input args.
    """
    with open(fname, 'r') as fin:
        lines = fin.readlines()
    line = lines[0]

    def read_fcidump(fd):
        with open(fd, 'r') as fin:
            lines = fin.readlines()
        ff = " ".join(lines).lower()
        pars = ff.split('/' if '/' in ff else '&end')[0]
        cont_dict = {}
        p_key = None
        for c in ','.join(pars.split()[1:]).split(','):
            if '=' in c or p_key is None:
                p_key, b = c.split('=')
                cont_dict[p_key.strip().lower()] = b.strip()
            elif len(c.strip()) != 0:
                if len(cont_dict[p_key.strip().lower()]) != 0:
                    cont_dict[p_key.strip().lower()] += ',' + c.strip()
                else:
                    cont_dict[p_key.strip().lower()] = c.strip()
        return cont_dict

    # construct input file from FCIDUMP
    if line.strip().lower().startswith('&fci '):
        cont_dict = read_fcidump(fname)
        lines = ["sym d2h", "orbitals %s" % fname,
                 "nelec %s" % cont_dict.get(
                     'nelec', 0), "spin %s" % cont_dict.get('ms2', 0),
                 "irrep %s" % cont_dict.get(
                     'isym', 0), "hf_occ integral", "schedule default",
                 "maxM 500", "maxiter 30"]

    dic = {}
    schedule = []
    site_dependent_bdims = []
    schedule_start = -1
    schedule_end = -1
    for i, line in enumerate(lines):
        if line.strip() != '' and "schedule" == line.strip().split()[0]:
            line_sp = line.strip().split()[1:]
            if len(line_sp) == 1:
                dic["schedule"] = line_sp[0]
            else:
                schedule_start = i
        elif "end" == line.strip():
            schedule_end = i
        elif schedule_start != -1 and schedule_end == -1 \
                and not line.strip().startswith('!') and not line.strip().startswith('#'):
            abcd = line.split()
            if len(abcd) == 4:
                a, b, c, d = abcd
                site_dependent_bdims.append([])
                schedule.append([int(a), int(b), float(c), float(d)])
            else:
                site_dependent_bdims.append([int(x) for x in abcd[1:-2]])
                schedule.append(
                    [int(abcd[0]), max(site_dependent_bdims[-1]), float(abcd[-2]), float(abcd[-1])])
        elif not line.strip().startswith('!') and not line.strip().startswith('#'):
            line_sp = line.split()
            if len(line_sp) != 0:
                if line_sp[0] in dic:
                    raise ValueError("duplicate key (%s)" % line_sp[0])
                dic[line_sp[0]] = " ".join(line_sp[1:])

    if len(schedule) == 0:
        schedule = get_schedule(dic)
        site_dependent_bdims = [[]] * len(schedule)

    tmp = list(zip(*schedule))
    nsweeps = np.diff(tmp[0]).tolist()
    maxiter = int(dic.get("maxiter", 1)) - int(np.sum(nsweeps))
    assert maxiter > 0
    nsweeps.append(maxiter)

    schedule = [[], [], [], []]
    for nswp, M, tol, noise, SM in zip(nsweeps, *tmp[1:], site_dependent_bdims):
        schedule[0].extend([M] * nswp)
        schedule[1].extend([tol] * nswp)
        schedule[2].extend([noise] * nswp)
        schedule[3].extend([SM] * nswp)
    dic["schedule"] = schedule

    # stopt extra keywords
    if "stopt_dmrg" in dic:
        dic["onepdm"] = ""
        dic["diag_twopdm"] = ""
        if "copy_mps" not in dic:
            dic["copy_mps"] = "ZKET"
        if "nonspinadapted" not in dic:
            dic["trans_mps_to_sz"] = ""
    if "stopt_compression" in dic:
        dic["compression"] = ""
        if "mps_tags" not in dic:
            dic["mps_tags"] = "BRA"
        if "read_mps_tags" not in dic:
            if "nonspinadapted" in dic:
                dic["read_mps_tags"] = "ZKET"
            else:
                dic["read_mps_tags"] = "KET"
        if "copy_mps" not in dic:
            dic["copy_mps"] = "ZBRA"
        if "nonspinadapted" not in dic:
            dic["trans_mps_to_sz"] = ""
    if "stopt_sampling" in dic:
        if "mps_tags" not in dic:
            if "nonspinadapted" in dic:
                dic["mps_tags"] = "ZKET ZBRA"
            else:
                dic["mps_tags"] = "KET BRA"

    # model extra keywords
    if "model" in dic:
        dic["noreorder"] = ""

    # optional keywords that can be obtained from fcidump
    if "orbitals" in dic and ("nelec" not in dic or "spin" not in dic or "irrep" not in dic
                              or ("k_symmetry" in dic and "k_irrep" not in dic)):
        if open(dic["orbitals"], 'rb').read(4) != b'\x89HDF':
            cont_dict = read_fcidump(dic["orbitals"])
            if "nelec" not in dic and "nelec" in cont_dict:
                dic["nelec"] = cont_dict["nelec"]
            if "spin" not in dic and "ms2" in cont_dict:
                dic["spin"] = cont_dict["ms2"]
            if "irrep" not in dic and "isym" in cont_dict:
                dic["irrep"] = cont_dict["isym"]
            if "k_symmetry" in dic and "k_irrep" not in dic and "kisym" in cont_dict:
                dic["k_irrep"] = cont_dict["kisym"]

    # sanity check
    diff = set(dic.keys()) - KNOWN_KEYS
    if len(diff) != 0:
        raise ValueError("Unrecognized keys (%s)" % diff)
    if "zerodot" in dic and ("twodot_to_onedot" in dic or "twodot" in dic):
        raise ValueError("zerodot conflicits with twodot_to_onedot/twodot.")
    if "onedot" in dic and ("twodot_to_onedot" in dic or "twodot" in dic):
        raise ValueError("onedot conflicits with twodot_to_onedot/twodot.")
    if "mem" in dic and (not dic["mem"][-1] in ['g', 'G']):
        raise ValueError("memory unit (%s) should be G" % (dic["mem"][-1]))
    if "intmem" in dic and (not dic["intmem"][-1] in ['g', 'G']):
        raise ValueError("memory unit (%s) should be G" % (dic["intmem"][-1]))
    crs = list(set(dic.keys()) & REORDER_KEYS)
    if len(crs) > 1:
        raise ValueError(
            "Reorder keys %s and %s cannot appear simultaneously." % (crs[0], crs[1]))
    crs = list(set(dic.keys()) & DYN_CORR_KEYS)
    if len(crs) > 1:
        raise ValueError(
            "Dynamic correlation keys %s and %s cannot appear simultaneously."
            % (crs[0], crs[1]))

    # restart check
    if "restart_oh" in dic:
        # OH is always fullrestart, and should not do dmrg or pdm
        dic.pop("onepdm", None)
        dic.pop("twopdm", None)
        dic.pop("correlation", None)
    if "diag_twopdm" in dic:
        dic["onepdm"] = ""
        dic["correlation"] = ""
    if "restart_diag_twopdm" in dic:
        dic["restart_onepdm"] = ""
        dic["restart_correlation"] = ""
    if len(set(dic.keys()) & RESTART_KEYS) != 0:
        dic["fullrestart"] = ""
    if len(set(dic.keys()) & NEVPT_KEYS) != 0:
        dic["onepdm"] = ""

    return dic


def parse_gaopt(fname):
    """
    parse a stackblock gaopt input file.

    Args:
        fname: stackblock gaopt input config file.

    Returns:
        dic: dictionary of input args.
    """
    with open(fname, 'r') as fin:
        lines = fin.readlines()
    dic = {}
    for line in lines:
        if not line.strip().startswith('!') and not line.strip().startswith('#'):
            line_sp = line.split()
            if len(line_sp) != 0:
                if line_sp[0] in dic:
                    raise ValueError("duplicate key (%s)" % line_sp[0])
                dic[line_sp[0]] = " ".join(line_sp[1:])

    # sanity check
    diff = set(dic.keys()) - GAOPT_KEYS
    if len(diff) != 0:
        raise ValueError("Unrecognized keys (%s)" % diff)

    return dic


def orbital_reorder(fcidump, method='gaopt'):
    """
    Find an optimal ordering of orbitals for DMRG.
    Ref: J. Chem. Phys. 142, 034102 (2015)

    Args:
        fcidump : block2.FCIDUMP object
        method :
            'gaopt <filename>/default' - genetic algorithm, take several seconds
            'fiedler' - very fast, may be slightly worse than 'gaopt'
            'manual <filename>' - manual reorder read from file
            'irrep <pointgroup>' - group same irrep together

    Return a index array "midx":
        reordered_orb_sym = original_orb_sym[midx]
    """
    from block2 import VectorDouble, OrbitalOrdering, VectorUInt16
    n_sites = fcidump.n_sites
    hmat = fcidump.abs_h1e_matrix()
    xmat = fcidump.abs_exchange_matrix()
    kmat = VectorDouble(np.array(hmat) * 1E-7 + np.array(xmat))
    if method.startswith("gaopt "):
        method = method[len("gaopt "):]
        dic = parse_gaopt(
            method) if method != '' and method != 'default' else {}
        assert dic.get("method", "gauss") == "gauss"
        assert float(dic.get("scale", 1.0)) == 1.0
        n_tasks = int(dic.get("maxcomm", 32))
        opts = dict(
            n_generations=int(dic.get("maxgen", 10000)),
            n_configs=int(dic.get("maxcell", n_sites * 2)),
            n_elite=int(dic.get("elite", 8)),
            clone_rate=1.0 - float(dic.get("cloning", 0.9)),
            mutate_rate=float(dic.get("mutation", 0.1))
        )
        midx, mf = None, None
        for _ in range(0, n_tasks):
            idx = OrbitalOrdering.ga_opt(n_sites, kmat, **opts)
            f = OrbitalOrdering.evaluate(n_sites, kmat, idx)
            idx = np.array(idx)
            if mf is None or f < mf:
                midx, mf = idx, f
    elif method == 'fiedler':
        idx = OrbitalOrdering.fiedler(n_sites, kmat)
        midx = np.array(idx)
    elif method.startswith("irrep "):
        method = method[len("irrep "):]
        if method == 'd2h':
            # D2H
            # 0   1   2   3   4   5   6   7   8   (FCIDUMP)
            #     A1g B3u B2u B1g B1u B2g B3g A1u
            # optimal
            #     A1g B1u B3u B2g B2u B3g B1g A1u
            optimal_reorder = [0, 1, 3, 5, 7, 2, 4, 6, 8]
        elif method == 'c2v':
            # C2V
            # 0  1  2  3  4  (FCIDUMP)
            #    A1 B1 B2 A2
            # optimal
            #    A1 B1 B2 A2
            optimal_reorder = [0, 1, 2, 3, 4]
        else:
            optimal_reorder = [0, 1, 3, 5, 7, 2, 4, 6, 8]
        orb_opt = [optimal_reorder[x] for x in np.array(fcidump.orb_sym)]
        idx = np.argsort(orb_opt)
        midx = np.array(idx)
    elif method.startswith("manual "):
        method = method[len("manual "):]
        idx = [int(x)
               for x in open(method[len("manual "):], "r").readline().split()]
        f = OrbitalOrdering.evaluate(n_sites, kmat, VectorUInt16(idx))
        idx = np.array(idx)
        midx, mf = idx, f
    else:
        raise ValueError("Unknown reorder method: %s" % method)
    fcidump.reorder(VectorUInt16(midx))
    return midx


def read_integral(fints, n_elec, twos, tol=1e-12, isym=1, orb_sym=None):
    """
    Read libDMET integral h5py file to block2 FCIDUMP object.

    Args:
        fints: h5 file
        n_elec: number of electrons
        twos: spin, nelec_a - nelec_b
        tol: tolerance of numerical zero
        isym: symmetry
        orb_sym: FCIDUMP orbital symm

    Returns:
        fcidump: block2 integral object.    
    """
    from libdmet_solid.system import integral
    from pyscf import ao2mo
    Ham = integral.load(fints)
    h1e = Ham.H1["cd"]
    g2e = Ham.H2["ccdd"]
    e_core = float(Ham.H0)
    n_sites = Ham.norb
    fcidump = FCIDUMP()

    if len(h1e) == 2:  # UHF
        mh1e_a = h1e[0][np.tril_indices(n_sites)]
        mh1e_a[np.abs(mh1e_a) < tol] = 0.0
        mh1e_b = h1e[1][np.tril_indices(n_sites)]
        mh1e_b[np.abs(mh1e_b) < tol] = 0.0
        mh1e = (mh1e_a, mh1e_b)

        g2e_aa = ao2mo.restore(8, g2e[0], n_sites)
        non0_idx = (g2e_aa < tol)
        non0_idx &= (g2e_aa > -tol)
        g2e_aa[non0_idx] = 0.0

        g2e_bb = ao2mo.restore(8, g2e[1], n_sites)
        non0_idx = (g2e_bb < tol)
        non0_idx &= (g2e_bb > -tol)
        g2e_bb[non0_idx] = 0.0

        g2e_ab = ao2mo.restore(4, g2e[2], n_sites)
        non0_idx = (g2e_ab < tol)
        non0_idx &= (g2e_ab > -tol)
        g2e_ab[non0_idx] = 0.0
        non0_idx = None

        mg2e = (g2e_aa, g2e_bb, g2e_ab)
        fcidump.initialize_sz(n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
    elif len(h1e) == 1:  # RHF
        mh1e = h1e[0][np.tril_indices(n_sites)]
        mh1e[np.abs(mh1e) < tol] = 0.0
        mg2e = ao2mo.restore(8, g2e[0], n_sites)
        non0_idx = (mg2e < tol)
        non0_idx &= (mg2e > -tol)
        mg2e[non0_idx] = 0.0
        non0_idx = None
        fcidump.initialize_su2(n_sites, n_elec, twos, isym, e_core, mh1e, mg2e)
    else:
        raise ValueError("h1e dimension error, len = %s", len(h1e))

    if orb_sym is None:
        orb_sym = [1] * n_sites
    fcidump.orb_sym = VectorUInt8(orb_sym)
    return fcidump


def format_schedule(sch):
    bdim, tol, noi, sbdim = sch
    if len(bdim) == 0:
        return ""
    lines = []
    for i in range(0, len(bdim)):
        if len(lines) != 0 and [bdim[i], noi[i], tol[i], sbdim[i]] == lines[-1][2:6]:
            lines[-1][1] = i
        else:
            lines.append([i, i, bdim[i], noi[i], tol[i], sbdim[i]])
    rr = []
    for l in lines:
        rr.append("Sweep%4d-%4d : Mmps = %5d Noise = %9.2g DavTol = %9.2g" % tuple(l[:-1]))
        if len(l[-1]) != 0:
            rr.append("--- SDP Mmps = %20s" % l[-1])
    return rr


def get_schedule(dic):
    start_m = int(dic.get("startM", 250))
    max_m = int(dic.get("maxM", 0))
    if max_m <= 0:
        raise ValueError("A positive maxM must be set for default schedule, " +
                         "current value : %d" % max_m)
    elif max_m < start_m:
        raise ValueError(
            "maxM %d cannot be smaller than startM %d" % (max_m, start_m))
    sch_type = dic.get("schedule", "")
    sweep_tol = float(dic.get("sweep_tol", 0))
    if sweep_tol == 0:
        dic["sweep_tol"] = "1E-5"
        sweep_tol = 1E-5
    schedule = []
    if sch_type == "default":
        def_m = [50, 100, 250, 500] + [1000 * x for x in range(1, 11)]
        def_iter = [8] * 5 + [4] * 9
        def_noise = [1E-3] * 3 + [1E-4] * 2 + [5E-5] * 9
        def_tol = [1E-4] * 3 + [1E-5] * 2 + [5E-6] * 9
        if start_m == max_m:
            schedule.append([0, start_m, 1E-5, 1E-4])
            schedule.append([8, start_m, 5E-6, 5E-5])
        elif start_m < def_m[0]:
            def_m.insert(0, start_m)
            for x in [def_iter, def_noise, def_tol]:
                x.insert(0, x[0])
        elif start_m > def_m[-1]:
            while start_m > def_m[-1]:
                def_m.append(def_m[-1] + 1000)
                for x in [def_iter, def_noise, def_tol]:
                    x.append(x[-1])
        else:
            for i in range(1, len(def_m)):
                if start_m < def_m[i]:
                    def_m[i - 1] = start_m
                    break
        isweep = 0
        for i in range(len(def_m)):
            if def_m[i] >= max_m:
                schedule.append([isweep, max_m, def_tol[i], def_noise[i]])
                isweep += def_iter[i]
                break
            elif def_m[i] >= start_m:
                schedule.append([isweep, def_m[i], def_tol[i], def_noise[i]])
                isweep += def_iter[i]
        schedule.append([schedule[-1][0] + 8, max_m, sweep_tol / 10, 0.0])
        last_iter = schedule[-1][0]
        if "twodot" not in dic and "onedot" not in dic and "twodot_to_onedot" not in dic:
            dic["twodot_to_onedot"] = str(last_iter + 2)
        max_iter = int(dic.get("maxiter", 0))
        if max_iter <= schedule[-1][0]:
            dic["maxiter"] = str(last_iter + 4)
            max_iter = last_iter + 4
    else:
        raise ValueError("Unrecognized schedule type (%s)" % sch_type)
    return schedule


def test_read_integral():
    # UHF intergal
    ints = read_integral("ints.h5", n_elec=4, twos=0,
                         tol=1e-12, isym=1, orb_sym=None)
    # RHF integral
    ints = read_integral("ints_res.h5", n_elec=16, twos=0,
                         tol=1e-8, isym=1, orb_sym=None)


if __name__ == "__main__":
    dic = parse(fname="./dmrg.conf")
    print(dic)
    test_read_integral()




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2021 Henrik R. Larsson <larsson@caltech.edu>
#  Copyright (C) 2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

""" Finite temperature DDMRG++ for Green's Function.

:author: Henrik R. Larsson, Sep 2021
        Based on zero temperature GFDMRG from Huanchen Zhai
"""

from block2 import SU2, SZ, Global, OpNamesSet, Threading, ThreadingTypes
from block2 import init_memory, release_memory, SiteIndex
from block2 import VectorUInt8, PointGroup, FCIDUMP, QCTypes, SeqTypes, OpNames, Random
from block2 import VectorUBond, VectorDouble, NoiseTypes, DecompositionTypes, EquationTypes
from block2 import OrbitalOrdering, VectorUInt16
from block2 import LinearSolverTypes

import time

# Set spin-adapted or non-spin-adapted here
#SpinLabel = SU2
SpinLabel = SZ

if SpinLabel == SU2:
    from block2.su2 import AncillaMPO, AncillaMPSInfo
    from block2.su2 import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.su2 import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO, IdentityAddedMPO
    from block2.su2 import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.su2 import VectorOpElement, LocalMPO
    try:
        from block2.su2 import MPICommunicator
        hasMPI = True
    except ImportError:
        hasMPI = False
    from ft_dmrg import FTDMRG_SU2 as FTDMRG
else:
    from block2.sz import AncillaMPO, AncillaMPSInfo
    from block2.sz import HamiltonianQC, SimplifiedMPO, Rule, RuleQC, MPOQC
    from block2.sz import MPSInfo, MPS, MovingEnvironment, DMRG, Linear, IdentityMPO, IdentityAddedMPO
    from block2.sz import OpElement, SiteMPO, NoTransposeRule, PDM1MPOQC, Expect
    from block2.sz import VectorOpElement, LocalMPO
    try:
        from block2.sz import MPICommunicator
        hasMPI = True
    except ImportError:
        hasMPI = False
    from ft_dmrg import FTDMRG_SZ as FTDMRG
import tools; tools.init(SpinLabel)
from tools import saveMPStoDir, loadMPSfromDir
import numpy as np
from typing import List, Union, Tuple


if hasMPI:
    MPI = MPICommunicator()
else:
    class _MPI:
        rank = 0
    MPI = _MPI()



_print = tools.getVerbosePrinter(MPI.rank == 0, flush=True)


class GFDMRG(FTDMRG):
    """
    DDMRG++ for Green's Function for molecules.
    """

    def greens_function(self, mps: MPS, E0: float, omegas: np.ndarray, eta: float,
                        idxs: List[int],
                        bond_dims: List[int], noises: List[float],
                        solver_tol: float, conv_tol: float, n_sweeps: int,
                        cps_bond_dims: List[int], cps_noises: List[float],
                        cps_conv_tol: float, cps_n_sweeps: float,
                        diag_only=False, alpha=True, addition = False,
                        cutoff=1E-14,
                        max_solver_iter=20000,
                        max_solver_iter_off_diag=0,
                        occs=None, bias=1.0, mo_coeff=None,
                        solver_type = LinearSolverTypes.LSQR,
                        use_preconditioner=False,
                        linear_solver_params = (80,-1),
                        callback=lambda i,j,w,gf:None) -> np.ndarray:
        """ Solve for the Green's function.
        GF_ij(omega + i eta) = <psi0| V_i' inv(H - E0 + omega + i eta) V_j |psi0>
        With V_i = a_i or a'_i.
        Note the definition of the sign of frequency omega.
        The GF is solved via the correction MPS C = inv(H - E0 + omega + i eta) V |psi0>

        :param mps: Start state psi0
        :param E0: Hamiltonian shift, typically ground state energy
        :param omegas: Frequencies the GF is computed
        :param eta: Broadening parameter
        :param idxs: GF orbital indices to compute GF_ij, with i,j in idxs
        :param bond_dims: Number of bond dimensions for each sweep for the correction MPS
        :param noises: Noises for each sweep
        :param solver_tol: Convergence tolerance for linear system solver
        :param conv_tol: Sweep convergence tolerance
        :param n_sweeps: Number of sweeps
        :param cps_bond_dims: Number of bond dimensions for each sweep for V |þsi0>
        :param cps_noises: Noises for each sweep for V |þsi0>
        :param cps_conv_tol:  Sweep convergence tolerance for V |þsi0>
        :param cps_n_sweeps: Number of sweeps for obtaining V |psi0>
        :param addition: If true, use -H + E0 instead of H - E0
        :param cutoff: Bond dimension cutoff for sweeps
        :param diag_only: Solve only diagonal of GF: GF_ii
        :param alpha: Creation/annihilation operator refers to alpha spin (otherwise: beta spin)
        :param max_solver_iter: Max. solver iterations for GF
        :param max_solver_iter_off_diag: Max. solver iteration for off diagonal terms (if 0, is set to max_solver_iter)
        :param occs: Optional occupation number vector for V|psi0> initialization
        :param bias: Optional occupation number bias for V|psi0> initialization
        :param mo_coeff: MPO is in MO basis but GF should be computed in AO basis
        :param solver_type: Linear solver type. Supported: GCROT; IDRS; LSQR
                LSQR is most robust
        :param use_preconditioner: Preconditioner for solver. No one or diagonal. No one is most robust.
        :param linear_solver_params: Sizes for either GCROT(M,K) or for IDR(S) (first entry; 2nd will be ignored)
            Will be ignored for LSQR.
        :param callback: Callback function after each GF computation.
                        Called as callback(i,j,w,GF_ij(omega))
        :return: the GF matrix
        """
        ops = [None] * len(idxs)
        rkets = [None] * len(idxs)
        rmpos = [None] * len(idxs)

        if self.mpi is not None:
            self.mpi.barrier()

        if self.mpo_orig is None:
            mpo = MPOQC(self.hamil, QCTypes.Conventional)
            mpo = SimplifiedMPO(AncillaMPO(mpo), RuleQC(), True, True,
                                OpNamesSet((OpNames.R, OpNames.RD)))
            self.mpo_orig = mpo


        if self.mpi is not None:
            if SpinLabel == SU2:
                from block2.su2 import ParallelMPO
            else:
                from block2.sz import ParallelMPO

        if addition:
            mpo = -1.0 * self.mpo_orig
            mpo.const_e += E0
        else:
            mpo = 1.0 * self.mpo_orig
            mpo.const_e -= E0

        #omegas += mpo.const_e
        #mpo.const_e = 0 # hrl: sometimes const_e causes trouble for AncillaMPO
        mpo = IdentityAddedMPO(mpo) # hrl: alternative

        if self.mpi is not None:
            mpo = ParallelMPO(mpo, self.prule)

        if self.print_statistics:
            _print('GF MPO BOND DIMS = ', ''.join(
                ["%6d" % (x.m * x.n) for x in mpo.left_operator_names]))
            max_d = max(bond_dims)
            mps_info2 = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                self.target, self.hamil.basis)
            mps_info2.set_bond_dimension(max_d)
            _, mem2, disk = mpo.estimate_storage(mps_info2, 2)
            _print("GF EST MAX MPS BOND DIMS = ", ''.join(
                ["%6d" % x.n_states_total for x in mps_info2.left_dims]))
            _print("GF EST PEAK MEM = ", GFDMRG.fmt_size(
                mem2), " SCRATCH = ", GFDMRG.fmt_size(disk))
            mps_info2.deallocate_mutable()
            mps_info2.deallocate()

        impo = SimplifiedMPO(AncillaMPO(IdentityMPO(self.hamil)),
                             NoTransposeRule(RuleQC()), True, True, OpNamesSet((OpNames.R, OpNames.RD)))

        if self.mpi is not None:
            impo = ParallelMPO(impo, self.identrule)

        def align_mps_center(ket, ref):
            # center itself may be "K" or "C"
            isOk = ket.center == ref.center
            isOk = isOk and ket.canonical_form[ket.center+1:] == ref.canonical_form[ref.center+1:]
            isOk = isOk and ket.canonical_form[:ket.center] == ref.canonical_form[:ref.center]
            if isOk:
                return
            if self.mpi is not None:
                self.mpi.barrier()
            cf = ket.canonical_form
            if ref.center == 0:
                ket.center += 1
                ket.canonical_form = ket.canonical_form[:-1] + 'S'
                while ket.center != 0:
                    ket.move_left(mpo.tf.opf.cg, self.prule)
            else:
                ket.canonical_form = 'K' + ket.canonical_form[1:]
                while ket.center != ket.n_sites - 1:
                    ket.move_right(mpo.tf.opf.cg, self.prule)
                ket.center -= 1
            if self.verbose >= 2:
                _print('CF = %s --> %s' % (cf, ket.canonical_form))

        ############################################
        # Prepare creation/annihilation operators
        ############################################
        if mo_coeff is None:
            if self.ridx is not None:
                gidxs = self.ridx[np.array(idxs)]
            else:
                gidxs = idxs
        else:
            if self.idx is not None:
                mo_coeff = mo_coeff[:, self.idx]
            gidxs = list(range(self.n_sites))
            ops = [None] * self.n_sites
            _print('idxs = ', idxs, 'gidxs = ', gidxs)

        for ii, idx in enumerate(gidxs):
            if SpinLabel == SZ:
                if addition:
                    ops[ii] = OpElement(OpNames.C, SiteIndex(
                        (idx, ), (0 if alpha else 1, )), SZ(1, 1 if alpha else -1, self.orb_sym[idx]))
                else:
                    ops[ii] = OpElement(OpNames.D, SiteIndex(
                        (idx, ), (0 if alpha else 1, )), SZ(-1, -1 if alpha else 1, self.orb_sym[idx]))
            else:
                if addition:
                    ops[ii] = OpElement(OpNames.C, SiteIndex(
                        (idx, ), ()), SU2(1, 1, self.orb_sym[idx]))
                else:
                    ops[ii] = OpElement(OpNames.D, SiteIndex(
                        (idx, ), ()), SU2(-1, 1, self.orb_sym[idx]))

        ############################################
        # Solve V_i |psi0>
        ############################################
        for ii, idx in enumerate(idxs):
            if self.mpi is not None:
                self.mpi.barrier()
            if self.verbose >= 2:
                _print('>>> START Compression Site = %4d <<<' % idx)
            t = time.perf_counter()

            rket_info = AncillaMPSInfo(self.n_physical_sites, self.hamil.vacuum,
                                       self.target + ops[ii].q_label, self.hamil.basis)
            rket_info.tag = 'DKET%d' % idx
            bond_dim = mps.info.bond_dim
            bond_dim = bond_dim if bond_dim != 0 else max([x.n_states_total for x in mps.info.left_dims])
            if occs is None:
                if self.verbose >= 2:
                    _print("Using FCI INIT MPS,bond_dim=", bond_dim)
                rket_info.set_bond_dimension(bond_dim)
            else:
                if self.verbose >= 2:
                    _print("Using occupation number INIT MPS; bond_dim=", bond_dim)
                rket_info.set_bond_dimension_using_occ(
                    bond_dim, VectorDouble(occs), bias=bias)
            rkets[ii] = MPS(self.n_sites, mps.center, 2)
            rkets[ii].initialize(rket_info)
            rkets[ii].random_canonicalize()

            rkets[ii].save_mutable()
            rkets[ii].deallocate()
            rket_info.save_mutable()
            rket_info.deallocate_mutable()

            if mo_coeff is None:
                # the mpo and gf are in the same basis
                # the mpo is SiteMPO
                rmpos[ii] = SimplifiedMPO(
                    AncillaMPO(SiteMPO(self.hamil, ops[ii])), NoTransposeRule(RuleQC()),
                    True, True, OpNamesSet((OpNames.R, OpNames.RD)))
            else:
                # the mpo is in mo basis and gf is in ao basis
                # the mpo is sum of SiteMPO (LocalMPO)
                ao_ops = VectorOpElement([None] * self.n_sites)
                for ix in range(self.n_physical_sites):
                    iix = ix * 2 # not ancilla sites
                    ao_ops[iix] = ops[iix] * mo_coeff[idx, ix]
                rmpos[ii] = SimplifiedMPO(
                    AncillaMPO(LocalMPO(self.hamil, ao_ops)), NoTransposeRule(RuleQC()),
                    True, True, OpNamesSet((OpNames.R, OpNames.RD)))

            if self.mpi is not None:
                rmpos[ii] = ParallelMPO(rmpos[ii], self.siterule)

            if len(cps_noises) == 1 and cps_noises[0] == 0:
                pme = None
            else:
                pme = MovingEnvironment(mpo, rkets[ii], rkets[ii], "PERT")
                pme.init_environments(False)
            rme = MovingEnvironment(rmpos[ii], rkets[ii], mps, "RHS")
            rme.init_environments(False)
            if self.delayed_contraction:
                if pme is not None:
                    pme.delayed_contraction = OpNamesSet.normal_ops()
                rme.delayed_contraction = OpNamesSet.normal_ops()

            # ME
            cps = Linear(pme, rme, VectorUBond(cps_bond_dims),
                         VectorUBond([mps.info.get_max_bond_dimension() + 100]),
                         VectorDouble(cps_noises))
            cps.noise_type = NoiseTypes.ReducedPerturbativeCollectedLowMem
            if pme is not None:
                cps.eq_type = EquationTypes.PerturbativeCompression
            cps.iprint = max(self.verbose - 1, 0)
            cps.cutoff = cutoff
            cps.solve(cps_n_sweeps, mps.center == 0, cps_conv_tol)


            if self.verbose >= 2:
                _print('>>> COMPLETE Compression Site = %4d | Time = %.2f <<<' %
                       (idx, time.perf_counter() - t))

        ############################################
        # Big GF LOOP
        ############################################
        gf_mat = np.zeros((len(idxs), len(idxs), len(omegas)), dtype=complex)

        for ii, idx in enumerate(idxs):

            if rkets[ii].center != mps.center:
                align_mps_center(rkets[ii], mps)
            lme = MovingEnvironment(mpo, rkets[ii], rkets[ii], "LHS")
            lme.init_environments(False)
            rme = MovingEnvironment(rmpos[ii], rkets[ii], mps, "RHS")
            rme.init_environments(False)
            if self.delayed_contraction:
                lme.delayed_contraction = OpNamesSet.normal_ops()
                rme.delayed_contraction = OpNamesSet.normal_ops()

            linear = Linear(lme, rme, VectorUBond(bond_dims),
                            VectorUBond(cps_bond_dims[-1:]), VectorDouble(noises))
            linear.gf_eta = eta
            linear.linear_conv_thrds = VectorDouble([solver_tol] * n_sweeps)
            linear.noise_type = NoiseTypes.ReducedPerturbativeCollectedLowMem
            linear.linear_solver_params = linear_solver_params
            linear.solver_type = solver_type
            linear.linear_use_precondition = use_preconditioner


            # TZ: Not raising error even if CG is not converged
            linear.linear_soft_max_iter = max_solver_iter
            linear.linear_max_iter = max_solver_iter + 1000
            linear.eq_type = EquationTypes.GreensFunction
            linear.iprint = max(self.verbose - 1, 0)
            linear.cutoff = cutoff

            for iw, w in enumerate(omegas):

                if self.verbose >= 2:
                    _print('>>>   START  GF OMEGA = %10.5f Site = %4d %4d <<<' %
                           (w, idx, idx))
                t = time.perf_counter()

                linear.tme = None
                linear.linear_soft_max_iter = max_solver_iter
                linear.noises = VectorDouble(noises)
                linear.bra_bond_dims = VectorUBond(bond_dims)
                linear.gf_omega = w
                linear.solve(n_sweeps, mps.center == 0, conv_tol)
                # set noises to 0 for next round
                noises = [0]
                min_site = np.argmin(np.array(linear.sweep_targets)[:, 1])
                if mps.center == 0:
                    min_site = mps.n_sites - 2 - min_site
                _print("GF.IMAG MIN SITE = %4d" % min_site)
                rgf, igf = linear.targets[-1]
                gf_mat[ii, ii, iw] = rgf + 1j * igf
                callback(ii,ii,w,gf_mat[ii,ii,iw])

                dmain, dseco, imain, iseco = Global.frame.peak_used_memory

                if self.verbose >= 1:
                    _print("=== %3s GF (%4d%4d | OMEGA = %10.5f ) = RE %20.15f + IM %20.15f === T = %7.2f" %
                           ("ADD" if addition else "REM", idx, idx, w, rgf, igf, time.perf_counter() - t))

                if self.verbose >= 2:
                    _print('>>> COMPLETE GF OMEGA = %10.5f Site = %4d %4d | Time = %.2f <<<' %
                           (w, idx, idx, time.perf_counter() - t))

                if diag_only:
                    continue

                for jj, idx2 in enumerate(idxs):

                    if jj > ii and rkets[jj].info.target == rkets[ii].info.target:

                        if rkets[jj].center != rkets[ii].center:
                            align_mps_center(rkets[jj], rkets[ii])

                        if self.verbose >= 2:
                            _print('>>>   START  GF OMEGA = %10.5f Site = %4d %4d <<<' % (
                                w, idx2, idx))
                        t = time.perf_counter()

                        tme = MovingEnvironment(
                            impo, rkets[jj], rkets[ii], "GF")
                        tme.init_environments(False)
                        if self.delayed_contraction:
                            tme.delayed_contraction = OpNamesSet.normal_ops()
                        linear.noises = VectorDouble(noises[-1:])
                        linear.bra_bond_dims = VectorUBond(bond_dims[-1:])
                        linear.target_bra_bond_dim = cps_bond_dims[-1]
                        linear.target_ket_bond_dim = cps_bond_dims[-1]
                        linear.tme = tme
                        if max_solver_iter_off_diag == 0:
                            linear.solve(1, mps.center != 0, 0)
                            rgf, igf = linear.targets[-1]
                        else:
                            linear.linear_soft_max_iter = max_solver_iter_off_diag if \
                                max_solver_iter_off_diag != -1 else max_solver_iter
                            linear.solve(1, mps.center == 0, 0)
                            if mps.center == 0:
                                rgf, igf = np.array(linear.sweep_targets)[::-1][min_site]
                            else:
                                rgf, igf = np.array(linear.sweep_targets)[min_site]
                        gf_mat[jj, ii, iw] = rgf + 1j * igf
                        gf_mat[ii, jj, iw] = rgf + 1j * igf
                        callback(ii,jj,w,gf_mat[ii,jj,iw])

                        if self.verbose >= 1:
                            _print("=== %3s GF (%4d%4d | OMEGA = %10.5f ) = RE %20.15f + IM %20.15f === T = %7.2f" %
                                   ("ADD" if addition else "REM", idx2, idx, w, rgf, igf, time.perf_counter() - t))

                        if self.verbose >= 2:
                            _print('>>> COMPLETE GF OMEGA = %10.5f Site = %4d %4d | Time = %.2f <<<' %
                                   (w, idx2, idx, time.perf_counter() - t))
        mps.save_data()
        mps.info.deallocate()

        if self.print_statistics:
            _print("GF PEAK MEM USAGE:",
                   "DMEM = ", GFDMRG.fmt_size(dmain + dseco),
                   "(%.0f%%)" % (dmain * 100 / (dmain + dseco)),
                   "IMEM = ", GFDMRG.fmt_size(imain + iseco),
                   "(%.0f%%)" % (imain * 100 / (imain + iseco)))

        return gf_mat

if __name__ == "__main__":

    # parameters
    n_threads = 2
    point_group = 'c1'
    scratch = '/tmp/block2'
    load_dir = None

    import os
    if MPI.rank == 0:
        tools.mkDir(scratch)
    if hasMPI:
        MPI.barrier()
    os.environ['TMPDIR'] = scratch
    E0 = -2.1379703474141984 #H4
    E0 = -1.568351864513 # H3


    MAX_M = 50
    gf_bond_dims=[MAX_M]
    gf_noises=[1E-3, 5E-4, 0]
    gf_tol=1E-4
    solver_tol=1E-6
    gf_n_steps = 20
    cps_bond_dims=[MAX_M] # Bond dimension of \hat a_i |psi> *and* |psi> (in time evolution)
    cps_noises=[0]
    cps_tol=1E-7
    cps_n_sweeps=30

    beta = 80 # inverse temperature
    dbeta = 0.5 # "time" step
    mu = -0.026282794560 # Chemical potential for initial state preparation

    #################################################
    # Prepare inital state
    #################################################

    dmrg = GFDMRG(scratch=scratch, memory=4e9,
                  verbose=3, omp_threads=n_threads)
    dmrg.init_hamiltonian_fcidump(point_group, "fcidump")
    #mps, mu = dmrg.optimize_mu(dmrg.fcidump.n_elec,mu, beta, dbeta, MAX_M)
    mps = dmrg.prepare_ground_state(mu, beta, dbeta, MAX_M)[0]


    eta = 0.005
    omegas = np.linspace(-1,0,200)
    # Frequencies more far away seem to be much harder to converge.
    # So it is better to start with small omegas and then use that as initial guess.
    omegas = omegas[::-1]
    idxs = [0] # Calc S_ii
    alpha = True # alpha or beta spin
    fOut = open("ft_gfdmrg_freqs.dat","w")
    fOut.write(f"# eta={eta}\n")
    fOut.write("# omega  Re(gf)  Im(gf)\n")
    def callback(i,j,omega,gf):
        print("-----",omega,":",gf)
        fOut.write(f"{omega:16.7f}   {gf.real:16.7f}  {gf.imag:16.7f}\n")
        fOut.flush()
    gf = dmrg.greens_function(mps, E0, omegas, eta, idxs,
                              gf_bond_dims, gf_noises, solver_tol, gf_tol, gf_n_steps,
                              cps_bond_dims, cps_noises, cps_tol, cps_n_sweeps,
                              addition=False, diag_only=False,
                              alpha=alpha, max_solver_iter_off_diag=0,
                              callback=callback)
    fOut.close()




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

"""
MPS/MPO algebra (fermionic, block-sparse).
"""

import numpy as np
from collections import Counter
from itertools import accumulate, groupby


class SubTensor:
    """
    A block in block-sparse tensor.

    Attributes:
        q_labels : tuple(SZ..)
            Quantum labels for this sub-tensor block.
            Each element in the tuple corresponds one rank of the tensor.
        reduced : numpy.ndarray
            Rank-:attr:`rank` dense reduced matrix.
        rank : int
            Rank of the tensor. ``rank == len(q_labels)``.
    """

    def __init__(self, q_labels=None, reduced=None):
        self.q_labels = tuple(q_labels) if q_labels is not None else ()
        self.rank = len(q_labels)
        self.reduced = reduced
        if self.rank != 0:
            if reduced is not None:
                assert len(self.reduced.shape) == self.rank

    def build_random(self):
        """Set reduced matrix with random numbers in [0, 1)."""
        self.reduced = np.random.random(self.reduced.shape)

    def build_zero(self):
        """Set reduced matrix to zero."""
        self.reduced = np.zeros(self.reduced.shape)
    
    def copy(self):
        """Shallow copy."""
        return SubTensor(q_labels=self.q_labels, reduced=self.reduced)

    def __mul__(self, other):
        """Scalar multiplication."""
        return SubTensor(q_labels=self.q_labels, reduced=other * self.reduced)

    def __neg__(self):
        """Times (-1)."""
        return SubTensor(q_labels=self.q_labels, reduced=-self.reduced)

    def equal_shape(self, other):
        """Test if two blocks have equal shape and quantum labels."""
        return self.q_labels == other.q_labels and self.reduced.shape == other.reduced.shape

    def __eq__(self, other):
        return self.q_labels == other.q_labels and np.allclose(self.reduced, other.reduced)

    def __repr__(self):
        return "(Q=) %r (R=) %r" % (self.q_labels, self.reduced)


class Tensor:
    """
    Block-sparse tensor.

    Attributes:
        blocks : list(SubTensor)
            A list of (non-zero) blocks.
    """

    def __init__(self, blocks=None):
        self.blocks = blocks if blocks is not None else []

    def copy(self):
        """Shallow copy."""
        return Tensor(blocks=self.blocks)

    def deep_copy(self):
        """Deep copy."""
        return Tensor(blocks=[b.copy() for b in self.blocks])

    def zero_copy(self):
        """A deep copy with zero reduced matrices."""
        blocks = [SubTensor(q_labels=b.q_labels, reduced=np.zeros_like(b.reduced))
                  for b in self.blocks]
        return Tensor(blocks=blocks)

    @property
    def rank(self):
        """Rank of the tensor."""
        return 0 if len(self.blocks) == 0 else self.blocks[0].rank

    @property
    def n_blocks(self):
        """Number of (non-zero) blocks."""
        return len(self.blocks)

    def modify(self, other):
        """Modify the blocks according to another Tensor's blocks."""
        self.blocks[:] = other.blocks

    def get_state_info(self, idx):
        """Get dict from quanta to number of states for left or right bond dimension."""
        mp = Counter()
        for block in self.blocks:
            q = block.q_labels[idx]
            if q in mp:
                assert block.reduced.shape[idx] == mp[q]
            else:
                mp[q] = block.reduced.shape[idx]
        return mp

    @staticmethod
    def contract(tsa, tsb, idxa, idxb, fidxa=None, fidxb=None, out_trans=None):
        """
        Contract two Tensor to form a new Tensor.

        Args:
            tsa : Tensor
                Tensor a, as left operand.
            tsb : Tensor
                Tensor b, as right operand.
            idxa : list(int)
                Indices of rank to be contracted in tensor a.
            idxb : list(int)
                Indices of rank to be contracted in tensor b.
            fidxa : None or int or (int, int)
                Index 'bra' in tensor a (operator), for determining fermion phase for 'op x st'.
                Or 'ket' in tensor a (operator), for determining fermion phase for 'op x op'.
                Or '(bra, bra)' in tensor a (operator product), for determining fermion phase for 'op x st'.
                This index should not be contracted.
                'fidxa' and 'fidxb' are only required for operator tensor applied on state tensor.
            fidxb : None or int or (int, int)
                Index 'virtual left' in tensor b (state), for determining fermion phase for 'op x st'.
                Or '(bra, ket)' in tensor b (operator), for determining fermion phase for 'op x op'.
                This index should not be contracted.
            out_trans : None or tuple
                Permutation of output indices.

        Returns:
            tensor : Tensor
                The contracted Tensor.
        """
        assert len(idxa) == len(idxb)
        idxa = [x if x >= 0 else tsa.rank + x for x in idxa]
        idxb = [x if x >= 0 else tsb.rank + x for x in idxb]
        out_idx_a = list(set(range(0, tsa.rank)) - set(idxa))
        out_idx_b = list(set(range(0, tsb.rank)) - set(idxb))

        # tuple of mutable object cannot be used as key
        map_idx_b = {}
        for block in tsb.blocks:
            subg = tuple(block.q_labels[id] for id in idxb)
            if subg not in map_idx_b:
                map_idx_b[subg] = []
            map_idx_b[subg].append(block)

        map_idx_out = {}
        for block_a in tsa.blocks:
            subg = tuple(block_a.q_labels[id] for id in idxa)
            if subg in map_idx_b:
                outga = tuple(block_a.q_labels[id] for id in out_idx_a)
                for block_b in map_idx_b[subg]:
                    outg = outga + \
                        tuple(block_b.q_labels[id] for id in out_idx_b)
                    outd = tuple(x for x in outg)
                    mat = np.tensordot(
                        block_a.reduced, block_b.reduced, axes=(idxa, idxb))
                    # fermionic phase factor
                    if isinstance(fidxa, tuple):
                        # operator product x two-site state
                        assert len(idxa) == 2 and len(fidxa) == 2
                        if (block_a.q_labels[fidxa[0]] + block_a.q_labels[idxa[0]] +
                            block_a.q_labels[fidxa[1]] + block_a.q_labels[idxa[1]]).is_fermion and \
                                block_b.q_labels[fidxb].is_fermion:
                            mat *= -1
                    elif fidxa is not None:
                        assert len(idxa) == 1
                        if isinstance(fidxb, tuple):
                            # operator x operator
                            if block_a.q_labels[fidxa].is_fermion and (
                                    block_b.q_labels[fidxb[0]] + block_b.q_labels[fidxb[1]]).is_fermion:
                                mat *= -1
                        else:
                            # operator x one-site state
                            if (block_a.q_labels[fidxa] + block_a.q_labels[idxa[0]]).is_fermion and \
                                    block_b.q_labels[fidxb].is_fermion:
                                mat *= -1
                    if out_trans is not None:
                        outg = tuple(outg[t] for t in out_trans)
                        mat = mat.transpose(out_trans)
                    if outd not in map_idx_out:
                        map_idx_out[outd] = SubTensor(
                            q_labels=outg, reduced=mat)
                    else:
                        map_idx_out[outd].reduced += mat
        if len(out_idx_a) + len(out_idx_b) == 0:
            if len(map_idx_out) == 0:
                return 0.0
            return map_idx_out[()].reduced.item()
        else:
            return Tensor(blocks=list(map_idx_out.values()))

    def left_canonicalize(self, mode='reduced'):
        """
        Left canonicalization (using QR factorization).
        Left canonicalization needs to collect all left indices for each specific right index.
        So that we will only have one R, but left dim of q is unchanged.

        Returns:
            r_blocks : dict(q_label_r -> numpy.ndarray)
                The R matrix for each right-index quantum label.
        """
        collected_rows = {}
        for block in self.blocks:
            q_label_r = block.q_labels[-1]
            if q_label_r not in collected_rows:
                collected_rows[q_label_r] = []
            collected_rows[q_label_r].append(block)
        r_blocks_map = {}
        for q_label_r, blocks in collected_rows.items():
            l_shapes = [np.prod(b.reduced.shape[:-1]) for b in blocks]
            mat = np.concatenate([b.reduced.reshape((sh, -1))
                                  for sh, b in zip(l_shapes, blocks)], axis=0)
            q, r = np.linalg.qr(mat, mode)
            r_blocks_map[q_label_r] = r
            qs = np.split(q, list(accumulate(l_shapes[:-1])), axis=0)
            assert(len(qs) == len(blocks))
            for q, b in zip(qs, blocks):
                b.reduced = q.reshape(b.reduced.shape[:-1] + (r.shape[0], ))
        return r_blocks_map

    def right_canonicalize(self, mode='reduced'):
        """
        Right canonicalization (using QR factorization).

        Returns:
            l_blocks : dict(q_label_l -> numpy.ndarray)
                The L matrix for each left-index quantum label.
        """
        collected_cols = {}
        for block in self.blocks:
            q_label_l = block.q_labels[0]
            if q_label_l not in collected_cols:
                collected_cols[q_label_l] = []
            collected_cols[q_label_l].append(block)
        l_blocks_map = {}
        for q_label_l, blocks in collected_cols.items():
            r_shapes = [np.prod(b.reduced.shape[1:]) for b in blocks]
            mat = np.concatenate([b.reduced.reshape((-1, sh)).T
                                  for sh, b in zip(r_shapes, blocks)], axis=0)
            q, r = np.linalg.qr(mat, mode)
            l_blocks_map[q_label_l] = r.T
            qs = np.split(q, list(accumulate(r_shapes[:-1])), axis=0)
            assert(len(qs) == len(blocks))
            for q, b in zip(qs, blocks):
                b.reduced = q.T.reshape((r.shape[0], ) + b.reduced.shape[1:])
        return l_blocks_map

    def left_multiply(self, mats):
        """
        Left Multiplication.
        Currently only used for multiplying R obtained from right-canonicalization/compression.

        Args:
            mats : dict(q_label_r -> numpy.ndarray)
                The R matrix for each right-index quantum label.
        """
        blocks = []
        for block in self.blocks:
            q_label_r = block.q_labels[0]
            if q_label_r in mats:
                mat = np.tensordot(
                    mats[q_label_r], block.reduced, axes=([1], [0]))
                blocks.append(SubTensor(q_labels=block.q_labels, reduced=mat))
        self.blocks = blocks

    def right_multiply(self, mats):
        """
        Right Multiplication.
        Currently only used for multiplying L obtained from right-canonicalization/compression.

        Args:
            mats : dict(q_label_l -> numpy.ndarray)
                The L matrix for each left-index quantum label.
        """
        blocks = []
        for block in self.blocks:
            q_label_l = block.q_labels[-1]
            if q_label_l in mats:
                mat = np.tensordot(
                    block.reduced, mats[q_label_l], axes=([block.rank - 1], [0]))
                blocks.append(SubTensor(q_labels=block.q_labels, reduced=mat))
        self.blocks = blocks

    def truncate_singular_values(self, svd_s, k=-1, cutoff=0.0):
        """
        Internal method for truncation.

        Args:
            svd_s : list(numpy.ndarray)
                Singular value array for each quantum number.
            k : int
                Maximal total bond dimension.
                If `k == -1`, no restriction in total bond dimension.
            cutoff : double
                Minimal kept singluar value.

        Returns:
            svd_r : list(numpy.ndarray)
                Truncated list of singular value arrays.
            gls : list(numpy.ndarray)
                List of kept singular value indices.
            error : double
                Truncation error (same unit as singular value).
        """
        ss = [(i, j, v) for i, ps in enumerate(svd_s)
              for j, v in enumerate(ps)]
        ss.sort(key=lambda x: -x[2])
        ss_trunc = [x for x in ss if x[2] >= cutoff]
        ss_trunc = ss_trunc[:k] if k != -1 else ss_trunc
        ss_trunc.sort(key=lambda x: (x[0], x[1]))
        svd_r = [None] * len(svd_s)
        gls = [None] * len(svd_s)
        error = 0.0
        for ik, g in groupby(ss_trunc, key=lambda x: x[0]):
            gl = np.array([ig[1] for ig in g], dtype=int)
            gl_inv = np.array(
                list(set(range(0, len(svd_s[ik]))) - set(gl)), dtype=int)
            gls[ik] = gl
            error += (svd_s[ik][gl_inv] ** 2).sum()
            svd_r[ik] = svd_s[ik][gl]
        for ik in range(len(svd_s)):
            if gls[ik] is None:
                error += (svd_s[ik] ** 2).sum()
        return svd_r, gls, np.sqrt(error)

    def left_compress(self, k=-1, cutoff=0.0):
        """
        Left compression needs to collect all left indices for each specific right index.
        Bond dimension of rightmost index is compressed.

        Args:
            k : int
                Maximal total bond dimension.
                If `k == -1`, no restriction in total bond dimension.
            cutoff : double
                Minimal kept singluar value.

        Returns:
            compressed tensor, dict of right blocks, compression error
        """
        collected_rows = {}
        for block in self.blocks:
            q_label_r = block.q_labels[-1]
            if q_label_r not in collected_rows:
                collected_rows[q_label_r] = []
            collected_rows[q_label_r].append(block)
        svd_s, blocks_l, blocks_r = [], [], []
        for q_label_r, blocks in collected_rows.items():
            l_shapes = [np.prod(b.reduced.shape[:-1]) for b in blocks]
            mat = np.concatenate([b.reduced.reshape((sh, -1))
                                  for sh, b in zip(l_shapes, blocks)], axis=0)
            u, s, vh = np.linalg.svd(mat, full_matrices=False)
            svd_s.append(s)
            blocks_l.append(u)
            blocks_r.append(vh)
        svd_r, gls, error = self.truncate_singular_values(svd_s, k, cutoff)
        for ik, gl in enumerate(gls):
            if gl is not None and len(gl) != len(svd_s[ik]):
                blocks_l[ik] = blocks_l[ik][:, gl]
                blocks_r[ik] = blocks_r[ik][gl, :]
        r_blocks_map = {}
        l_blocks = []
        for ik, (q_label_r, blocks) in enumerate(collected_rows.items()):
            if svd_r[ik] is not None:
                l_shapes = [np.prod(b.reduced.shape[:-1]) for b in blocks]
                qs = np.split(blocks_l[ik], list(
                    accumulate(l_shapes[:-1])), axis=0)
                for q, b in zip(qs, blocks):
                    mat = q.reshape(
                        b.reduced.shape[:-1] + (blocks_l[ik].shape[1], ))
                    l_blocks.append(
                        SubTensor(q_labels=b.q_labels, reduced=mat))
                r_blocks_map[q_label_r] = svd_r[ik][:, None] * blocks_r[ik]
        return Tensor(blocks=l_blocks), r_blocks_map, error

    def right_compress(self, k=-1, cutoff=0.0):
        """
        Right compression needs to collect all right indices for each specific left index.
        Bond dimension of leftmost index is compressed.

        Args:
            k : int
                Maximal total bond dimension.
                If `k == -1`, no restriction in total bond dimension.
            cutoff : double
                Minimal kept singluar value.

        Returns:
            compressed tensor, dict of left blocks, compression error
        """
        collected_cols = {}
        for block in self.blocks:
            q_label_l = block.q_labels[0]
            if q_label_l not in collected_cols:
                collected_cols[q_label_l] = []
            collected_cols[q_label_l].append(block)
        svd_s, blocks_l, blocks_r = [], [], []
        for q_label_l, blocks in collected_cols.items():
            r_shapes = [np.prod(b.reduced.shape[1:]) for b in blocks]
            mat = np.concatenate([b.reduced.reshape((-1, sh))
                                  for sh, b in zip(r_shapes, blocks)], axis=1)
            u, s, vh = np.linalg.svd(mat, full_matrices=False)
            svd_s.append(s)
            blocks_l.append(u)
            blocks_r.append(vh)
        svd_r, gls, error = self.truncate_singular_values(svd_s, k, cutoff)
        for ik, gl in enumerate(gls):
            if gl is not None and len(gl) != len(svd_s[ik]):
                blocks_l[ik] = blocks_l[ik][:, gl]
                blocks_r[ik] = blocks_r[ik][gl, :]
        l_blocks_map = {}
        r_blocks = []
        for ik, (q_label_l, blocks) in enumerate(collected_cols.items()):
            if svd_r[ik] is not None:
                r_shapes = [np.prod(b.reduced.shape[1:]) for b in blocks]
                qs = np.split(blocks_r[ik], list(
                    accumulate(r_shapes[:-1])), axis=1)
                assert(len(qs) == len(blocks))
                for q, b in zip(qs, blocks):
                    mat = q.reshape(
                        (blocks_r[ik].shape[0], ) + b.reduced.shape[1:])
                    r_blocks.append(
                        SubTensor(q_labels=b.q_labels, reduced=mat))
                l_blocks_map[q_label_l] = svd_r[ik][None, :] * blocks_l[ik]
        return Tensor(blocks=r_blocks), l_blocks_map, error

    def __mul__(self, other):
        """Scalar multiplication."""
        return Tensor(blocks=[block * other for block in self.blocks])

    def __neg__(self):
        """Times (-1)."""
        return Tensor(blocks=[-block for block in self.blocks])

    def __repr__(self):
        return "\n".join("%3d %r" % (ib, b) for ib, b in enumerate(self.blocks))


class MPS:
    """
    Matrix Product State.

    Attributes:
        tensors : list(Tensor)
            A list of MPS tensors.
        n_sites : int
            Number of sites.
    """

    def __init__(self, tensors=None):
        self.tensors = tensors if tensors is not None else []

    @property
    def n_sites(self):
        """Number of sites."""
        return len(self.tensors)

    def deep_copy(self):
        """Deep copy."""
        return MPS(tensors=[ts.deep_copy() if ts is not None else None for ts in self.tensors])

    def get_left_dims(self, idx=0):
        """
        Get list of dict from quanta to number of states for left bond dimension of each site tensor.

        Args:
            idx : int
                Index of left bond.
        """
        left_dims = []
        for i in range(1, self.n_sites):
            if self.tensors[i] is None:
                left_dims.append(Counter())
            else:
                left_dims.append(self.tensors[i].get_state_info(idx))
        return left_dims

    def get_right_dims(self, idx=-1):
        """
        Get list of dict from quanta to number of states for right bond dimension of each site tensor.

        Args:
            idx : int
                Index of right bond.
        """
        right_dims = []
        for i in range(self.n_sites - 1):
            if self.tensors[i] is None:
                right_dims.append(Counter())
            else:
                right_dims.append(self.tensors[i].get_state_info(idx))
        return right_dims

    def get_bond_dims(self, idxl=0, idxr=-1):
        """
        Get list of dict from quanta to number of states for bond dimension of each site tensor.
        This is the union of left and right dimensions at each bond.

        Args:
            idxl : int
                Index of left bond.
            idxr : int
                Index of right bond (can be negative).
        """
        left_dims = self.get_left_dims(idxl)
        right_dims = self.get_right_dims(idxr)
        bond_dims = []
        for ld, rd in zip(left_dims, right_dims):
            bond_dims.append(ld | rd)
        return bond_dims

    def __mul__(self, other):
        """Scalar multiplication."""
        return MPS(tensors=[self.tensors[0] * other] + self.tensors[1:])

    def __rmul__(self, other):
        return self * other

    def __neg__(self):
        """Times (-1)."""
        return MPS(tensors=[-self.tensors[0]] + self.tensors[1:])

    def __add__(self, other):
        """Add two MPS. data in `other` MPS will be put in larger reduced indices."""
        assert isinstance(other, MPS)
        assert self.n_sites == other.n_sites

        bonds = self.get_bond_dims(), other.get_bond_dims()
        sum_bonds = [bondx + bondy for bondx, bondy in zip(*bonds)]

        tensors = []
        for i in range(self.n_sites):
            if self.tensors[i] is None:
                assert other.tensors[i] is None
                tensors.append(None)
                continue
            if i != 0:
                lb = sum_bonds[i - 1]
            if i != self.n_sites - 1:
                rb = sum_bonds[i]
            sub_mp = {}
            # find required new blocks and their shapes
            for block in self.tensors[i].blocks + other.tensors[i].blocks:
                q = block.q_labels
                sh = block.reduced.shape
                if q not in sub_mp:
                    mshape = list(sh)
                    if i != 0:
                        mshape[0] = lb[q[0]]
                    if i != self.n_sites - 1:
                        mshape[-1] = rb[q[-1]]
                    sub_mp[q] = SubTensor(q, np.zeros(tuple(mshape)))
            # copy block self.blocks to smaller index in new block
            for block in self.tensors[i].blocks:
                q = block.q_labels
                sh = block.reduced.shape
                if i == 0:
                    sub_mp[q].reduced[..., : sh[-1]] += block.reduced
                elif i == self.n_sites - 1:
                    sub_mp[q].reduced[: sh[0], ...] += block.reduced
                else:
                    sub_mp[q].reduced[: sh[0], ..., : sh[-1]] += block.reduced
            # copy block other.blocks to greater index in new block
            for block in other.tensors[i].blocks:
                q = block.q_labels
                sh = block.reduced.shape
                if i == 0:
                    sub_mp[q].reduced[..., -sh[-1]:] += block.reduced
                elif i == self.n_sites - 1:
                    sub_mp[q].reduced[-sh[0]:, ...] += block.reduced
                else:
                    sub_mp[q].reduced[-sh[0]:, ..., -sh[-1]:] += block.reduced
            tensors.append(Tensor(blocks=list(sub_mp.values())))
        return MPS(tensors=tensors)

    def __sub__(self, other):
        return self + (-other)

    def __or__(self, other):
        """
        Contraction of two general MPS to a number. <MPS|MPS>.
        A general MPS is MPS/MPO with one or more physical indices in the middle,
        but two-site tensor is not allowed.
        """

        assert self.__class__ == other.__class__
        assert self.n_sites == other.n_sites
        left = 0.0
        for i in range(self.n_sites):
            assert self.tensors[i] is not None
            assert other.tensors[i] is not None
            assert self.tensors[i].rank == other.tensors[i].rank
            if self.tensors[i].n_blocks == 0 or other.tensors[i].n_blocks == 0:
                return 0.0
            if i != self.n_sites - 1:
                cidx = list(range(0, self.tensors[i].rank - 1))
            else:
                cidx = list(range(0, self.tensors[i].rank))
            if i == 0:
                left = Tensor.contract(self.tensors[i], other.tensors[i], cidx, cidx)
            else:
                lbra = Tensor.contract(left, self.tensors[i], [0], [0])
                left = Tensor.contract(lbra, other.tensors[i], cidx, cidx)
        assert isinstance(left, float)

        return left

    def __matmul__(self, other):
        """Contraction of two MPS. <MPS|MPS>."""

        assert isinstance(other, MPS)
        if isinstance(other, MPO):
            return other @ self
        assert self.n_sites == other.n_sites

        left = 0.0
        for i in range(self.n_sites):
            if self.tensors[i] is None:
                assert other.tensors[i] is None
                continue
            if self.tensors[i].n_blocks == 0 or other.tensors[i].n_blocks == 0:
                continue
            if i == 0:
                if self.tensors[i].rank == 3 and self.tensors[i + 1] is None:
                    left = Tensor.contract(
                        self.tensors[i], other.tensors[i], [0, 1], [0, 1])
                else:
                    left = Tensor.contract(
                        self.tensors[i], other.tensors[i], [0], [0])
            else:
                lbra = Tensor.contract(left, self.tensors[i], [0], [0])
                if self.tensors[i].rank == 4 or (
                        i == self.n_sites - 2 and self.tensors[i + 1]
                        is None and self.tensors[i].rank == 3):
                    left = Tensor.contract(lbra, other.tensors[i], [
                                           0, 1, 2], [0, 1, 2])
                else:
                    left = Tensor.contract(
                        lbra, other.tensors[i], [0, 1], [0, 1])
        assert isinstance(left, float)

        return left

    def canonicalize(self, center):
        """
        MPS/MPO canonicalization.

        Args:
            center : int
                Site index of canonicalization center.
        """
        for i in range(0, center):
            if self.tensors[i] is None:
                continue
            rs = self.tensors[i].left_canonicalize()
            if i + 1 < self.n_sites and self.tensors[i + 1] is not None:
                self.tensors[i + 1].left_multiply(rs)
            elif i + 2 < self.n_sites:
                self.tensors[i + 2].left_multiply(rs)
        for i in range(self.n_sites - 1, center, -1):
            if self.tensors[i] is None:
                continue
            ls = self.tensors[i].right_canonicalize()
            if i - 1 >= 0 and self.tensors[i - 1] is not None:
                self.tensors[i - 1].right_multiply(ls)
            elif i - 2 >= 0:
                self.tensors[i - 2].right_multiply(ls)

    def compress(self, k=-1, cutoff=0.0, left=True):
        """
        MPS/MPO bond dimension compression.

        Args:
            k : int
                Maximal total bond dimension.
                If `k == -1`, no restriction in total bond dimension.
            cutoff : double
                Minimal kept singluar value.
            left : bool
                If left, canonicalize to right boundary and then svd to left.
                Otherwise, canonicalize to left boundary and then svd to right.
        """
        merror = 0.0
        if left:
            self.canonicalize(self.n_sites - 1)
            for i in range(self.n_sites - 1, 0, -1):
                if self.tensors[i] is None:
                    continue
                r, ls, err = self.tensors[i].right_compress(k, cutoff)
                merror = max(merror, err)
                self.tensors[i] = r
                if self.tensors[i - 1] is not None:
                    self.tensors[i - 1].right_multiply(ls)
                elif i - 2 >= 0:
                    self.tensors[i - 2].right_multiply(ls)
        else:
            self.canonicalize(0)
            for i in range(0, self.n_sites - 1):
                if self.tensors[i] is None:
                    continue
                l, rs, err = self.tensors[i].left_compress(k, cutoff)
                merror = max(merror, err)
                self.tensors[i] = l
                if self.tensors[i + 1] is not None:
                    self.tensors[i + 1].left_multiply(rs)
                elif i + 2 < self.n_sites:
                    self.tensors[i + 2].left_multiply(rs)
        return merror

    def merge_virtual_dims(self):
        """Merge double left and right virtual dims to single left and right virtual dim."""
        bonds_up = self.get_bond_dims(idxl=0, idxr=-2)
        bonds_dn = self.get_bond_dims(idxl=1, idxr=-1)

        # tensor product of virtual dimensions
        bonds_maps = []
        for bup, bdn in zip(bonds_up, bonds_dn):
            bonds_map = {}
            for u, nu in bup.items():
                for d, nd in bdn.items():
                    q = u + d
                    if q not in bonds_map:
                        bonds_map[q] = [0, {}]
                    assert (u, d) not in bonds_map[q][1]
                    bonds_map[q][1][(u, d)] = bonds_map[q][0]
                    bonds_map[q][0] += nu * nd
            bonds_maps.append(bonds_map)

        # merge virtual dimensions
        for i in range(self.n_sites):
            if self.tensors[i] is None:
                continue
            map_blocks = {}
            for block in self.tensors[i].blocks:
                if i != 0:
                    qla = block.q_labels[0]
                    qlb = block.q_labels[1]
                    ql = qla + qlb
                    nl = bonds_maps[i - 1][ql][0]
                    ill = bonds_maps[i - 1][ql][1][(qla, qlb)]
                    nll = block.reduced.shape[0] * block.reduced.shape[1]
                if i != self.n_sites - 1:
                    qra = block.q_labels[-2]
                    qrb = block.q_labels[-1]
                    qr = qra + qrb
                    nr = bonds_maps[i][qr][0]
                    irr = bonds_maps[i][qr][1][(qra, qrb)]
                    nrr = block.reduced.shape[-2] * block.reduced.shape[-1]
                if i == 0:
                    qm = block.q_labels[:-2]
                    q = qm + (qr,)
                    sh = block.reduced.shape[:-2] + (nr,)
                    xsh = block.reduced.shape[:-2] + (nrr,)
                    if q not in map_blocks:
                        map_blocks[q] = SubTensor(
                            q_labels=q, reduced=np.zeros(sh))
                    map_blocks[q].reduced[..., irr: irr
                                          + nrr] = block.reduced.reshape(xsh)
                elif i == self.n_sites - 1:
                    qm = block.q_labels[2:]
                    q = (ql, ) + qm
                    sh = (nl,) + block.reduced.shape[2:]
                    xsh = (nll,) + block.reduced.shape[2:]
                    if q not in map_blocks:
                        map_blocks[q] = SubTensor(
                            q_labels=q, reduced=np.zeros(sh))
                    map_blocks[q].reduced[ill:ill
                                          + nll, ...] = block.reduced.reshape(xsh)
                else:
                    qm = block.q_labels[2:-2]
                    q = (ql, ) + qm + (qr, )
                    sh = (nl,) + block.reduced.shape[2:-2] + (nr,)
                    xsh = (nll,) + block.reduced.shape[2:-2] + (nrr,)
                    if q not in map_blocks:
                        map_blocks[q] = SubTensor(
                            q_labels=q, reduced=np.zeros(sh))
                    map_blocks[q].reduced[ill:ill + nll, ...,
                                          irr:irr + nrr] = block.reduced.reshape(xsh)
            self.tensors[i] = Tensor(blocks=list(map_blocks.values()))

    def show_bond_dims(self):
        bonds = self.get_bond_dims(idxl=0, idxr=-1)
        return '|'.join([str(sum(x.values())) for x in bonds])

    def __getitem__(self, i):
        return self.tensors[i]

    def __setitem__(self, i, ts):
        self.tensors[i] = ts

    def norm(self):
        return np.sqrt(self @ self)


class MPO(MPS):
    """
    Matrix Product Operator.

    Attributes:
        tensors : list(Tensor)
            A list of MPO tensors.
        const_e : float
            constant energy term.
        n_sites : int
            Number of sites.
    """

    def __init__(self, tensors=None, const_e=0.0):
        self.const_e = const_e
        super().__init__(tensors=tensors)

    def deep_copy(self):
        """Deep copy."""
        return MPO(tensors=[ts.deep_copy() for ts in self.tensors], const_e=self.const_e)

    def __mul__(self, other):
        """Scalar multiplication."""
        return MPO(tensors=[self.tensors[0] * other] + self.tensors[1:], const_e=other * self.const_e)


    def __neg__(self):
        """Times (-1)."""
        return MPO(tensors=[-self.tensors[0]] + self.tensors[1:], const_e=-self.const_e)

    def __add__(self, other):
        """Add two MPO. data in `other` MPO will be put in larger reduced indices."""
        assert isinstance(other, MPO)
        return MPO(tensors=super().__add__(other).tensors, const_e=self.const_e + other.const_e)

    def __matmul__(self, other):
        """
        (a) Contraction of MPO and MPS. MPO |MPS>. (other : MPS)
        (b) Contraction of MPO and MPO. MPO * MPO. (other : MPO)
        """

        tensors = [None] * self.n_sites
        if isinstance(other, MPO):
            assert self.n_sites == other.n_sites

            for i in range(self.n_sites):
                if i == 0:
                    ot = (0, 2, 1, 3)
                    tensors[i] = Tensor.contract(
                        self.tensors[i], other.tensors[i], [1], [0], out_trans=ot)
                else:
                    ot = (0, 2, 1, 3) if i == self.n_sites - \
                        1 else (0, 3, 1, 4, 2, 5)
                    tensors[i] = Tensor.contract(
                        self.tensors[i], other.tensors[i],
                        [2], [1], 1, 0, out_trans=ot)

            mpo = MPO(tensors=tensors)
            mpo.merge_virtual_dims()
            if self.const_e == 0 and other.const_e == 0:
                return mpo
            elif self.const_e == 0:
                return mpo + self * other.const_e
            elif other.const_e == 0:
                return mpo + other * self.const_e
            else:
                mpo = mpo + (other * self.const_e + self * other.const_e)
                mpo.const_e -= self.const_e * other.const_e
                return mpo
        elif isinstance(other, MPS):
            assert self.n_sites == other.n_sites

            for i in range(self.n_sites):
                if other.tensors[i] is None:
                    continue
                if i == 0:
                    if i < self.n_sites - 1 and other.tensors[i + 1] is None:
                        # 2-site case (first tensor or the only tensor)
                        op = Tensor.contract(
                            self.tensors[i], self.tensors[i + 1], [2], [0], 1, (1, 2))
                        tensors[i] = Tensor.contract(
                            op, other.tensors[i], [1, 3], [0, 1])
                    else:
                        # 1-site case (first tensor)
                        tensors[i] = Tensor.contract(
                            self.tensors[i], other.tensors[i], [1], [0])
                else:
                    if other.tensors[i].rank == 4 or \
                            (i == self.n_sites - 2 and other.tensors[i + 1] is None and other.tensors[i].rank == 3):
                        if other.tensors[i].rank == 4:
                            assert i < self.n_sites - 1
                            assert other.tensors[i + 1] is None
                            ot = (0, 4, 1, 2, 3, 5)
                        else:
                            ot = (0, 3, 1, 2)
                        # 2-site case
                        op = Tensor.contract(
                            self.tensors[i], self.tensors[i + 1], [3], [0], 2, (1, 2))
                        tensors[i] = Tensor.contract(op, other.tensors[i],
                                                     [2, 4], [1, 2], (1, 3), 0, out_trans=ot)
                    else:
                        # 1-site case
                        ot = (0, 2, 1) if i == self.n_sites - \
                            1 else (0, 3, 1, 2, 4)
                        tensors[i] = Tensor.contract(
                            self.tensors[i], other.tensors[i],
                            [2], [1], 1, 0, out_trans=ot)

            mps = MPS(tensors=tensors)
            mps.merge_virtual_dims()
            return mps + self.const_e * other if self.const_e != 0 else mps
        else:
            assert False




#  block2: Efficient MPO implementation of quantum chemistry DMRG
#  Copyright (C) 2020-2021 Huanchen Zhai <hczhai@caltech.edu>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#

import block2
from block2 import *
from block2.su2 import *
from pyscf import lib, ao2mo, gto, mcscf, tools, symm
import numpy as np

try:
    from pyblock2.dmrgscf import DMRGCI as _DMRGCI
except ImportError:
    from dmrgscf import DMRGCI as _DMRGCI

class NEVPT(lib.StreamObject):

    def __init__(self, mc, mp_order=2, ci_order=2):
        self._casci = mc
        self._scf = mc._scf
        self.mol = self._scf.mol
        self.verbose = self.mol.verbose
        self.stdout = self.mol.stdout
        self.e_corr = None
        self.mp_order = mp_order
        self.ci_order = ci_order
        self.dmrg_args = {
            "startM": 250, "maxM": 500, "schedule": "default",
            "sweep_tol": 1E-6, "cutoff": 1E-14,
            "memory": lib.param.MAX_MEMORY * 1E6
        }

    @property
    def e_tot(self):
        return np.asarray(self.e_corr) + self._scf.e_tot

    def kernel(self, mo_coeff=None):

        if mo_coeff is None:
            mo_coeff = self._scf.mo_coeff

        # Global
        Random.rand_seed(123456)
        scratch = './nodex'
        n_threads = lib.num_threads()
        fcidump_tol = 1E-13
        memory = self.dmrg_args["memory"]
        init_memory(isize=int(memory * 0.05),
            dsize=int(memory * 0.95), save_dir=scratch)
        Global.threading = Threading(
            ThreadingTypes.OperatorBatchedGEMM | ThreadingTypes.Global,
            n_threads, n_threads, 1)
        Global.threading.seq_type = SeqTypes.Nothing
        Global.frame.fp_codec = DoubleFPCodec(1E-16, 1024)
        Global.frame.load_buffering = False
        Global.frame.save_buffering = False
        Global.frame.use_main_stack = False
        Global.frame.minimal_disk_usage = True
        if self.verbose >= 5:
            print(Global.frame)
            print(Global.threading)
        
        mc = self._casci
        
        if self.mol.groupname.lower() == 'c1':
            orb_sym = [0] * self.mol.nao
        else:
            orb_sym = irrep_name2id(symm.label_orb_symm(
                self.mol, self.mol.irrep_name, self.mol.symm_orb, mo_coeff))

        # FCIDUMP
        h1e = mo_coeff.T @ self._scf.get_hcore() @ mo_coeff
        e_core = self.mol.energy_nuc()
        h1e = h1e.flatten()
        g2e = ao2mo.restore(8, ao2mo.kernel(self.mol, mo_coeff), self.mol.nao)
        h1e[np.abs(h1e) < fcidump_tol] = 0
        g2e[np.abs(g2e) < fcidump_tol] = 0
        na, nb = self.mol.nelec
        fcidump = FCIDUMP()
        fcidump.initialize_su2(self.mol.nao, na + nb, abs(na - nb), 1, e_core, h1e, g2e)
        error = fcidump.symmetrize(VectorUInt8(orb_sym))
        if self.verbose >= 5:
            print('symm error = ', error)

        n_orbs = self.mol.nao
        self.n_orbs = self.mol.nao

        # Hamiltonian
        n_inactive = mc.ncore
        n_external = self.n_orbs - mc.ncas - mc.ncore
        assert n_inactive + n_external <= n_orbs
        big_left_orig = CSFBigSite(n_inactive, self.ci_order, False,
            fcidump, VectorUInt8(orb_sym[:n_inactive]), max(min(self.verbose - 4, 3), 0))
        big_right_orig = CSFBigSite(n_external, self.ci_order, True,
            fcidump, VectorUInt8(orb_sym[-n_external:]), max(min(self.verbose - 4, 3), 0))
        big_left = SimplifiedBigSite(big_left_orig, RuleQC())
        big_right = SimplifiedBigSite(big_right_orig, RuleQC())
        vacuum = SU2(0)
        target = SU2(na + nb, abs(na - nb), 0)
        hamil = HamiltonianQCBigSite(vacuum, n_orbs, VectorUInt8(orb_sym), fcidump,
            big_left, big_right)
        n_sites = hamil.n_sites

        self.hamil = hamil

        if self.dmrg_args["schedule"] == "default":
            _DMRGCI.get_schedule(self.dmrg_args)

        bond_dims, dav_thrds, noises = self.dmrg_args["schedule"]
        sweep_tol = self.dmrg_args["sweep_tol"]

        # MPS
        info = CASCIMPSInfo(n_sites, vacuum, target, hamil.basis,
            1 if n_inactive != 0 else 0, mc.ncas, 1 if n_external != 0 else 0)
        info.set_bond_dimension(bond_dims[0])
        mps = MPS(n_sites, 0, 2)
        mps.initialize(info)
        mps.random_canonicalize()
        mps.tensors[mps.center].normalize()
        mps.save_mutable()
        info.save_mutable()
        forward = mps.center == 0

        # MPO
        mpo = MPOQC(hamil, QCTypes.NC)
        mpo = SimplifiedMPO(mpo, RuleQC(), True, True, OpNamesSet((OpNames.R, OpNames.RD)))

        # DMRG
        me = MovingEnvironment(mpo, mps, mps, "DMRG")
        me.delayed_contraction = OpNamesSet.normal_ops()
        me.cached_contraction = True
        me.save_partition_info = True
        me.init_environments(self.verbose >= 5)
        dmrg = DMRGBigSite(me, VectorUBond(bond_dims), VectorDouble(noises))
        dmrg.last_site_svd = True
        dmrg.last_site_1site = True
        dmrg.decomp_last_site = False
        dmrg.davidson_conv_thrds = VectorDouble(dav_thrds)
        dmrg.cutoff = self.dmrg_args["cutoff"]
        dmrg.iprint = max(min(self.verbose - 4, 3), 0)
        dmrg.noise_type = NoiseTypes.ReducedPerturbativeCollected
        ener = dmrg.solve(len(bond_dims), forward, sweep_tol)

        self.converged = True
        self.e_corr = ener - self._scf.e_tot

        lib.logger.note(self, 'E(NEVPT1) = %.16g  E_corr = %.16g', self.e_tot, self.e_corr)

        self.mps = mps

        if self.mp_order == 1:
            return self.e_corr, mps

        # FCIDUMP
        dm1 = self.make_rdm1(mps).copy(order='C')
        fd_dyall = DyallFCIDUMP(fcidump, n_inactive, n_external)
        fd_dyall.initialize_from_1pdm_su2(dm1)
        error = fd_dyall.symmetrize(VectorUInt8(orb_sym))
        if self.verbose >= 5:
            print('symm error = ', error)
        
        # Hamiltonian
        big_left = SimplifiedBigSite(big_left_orig, NoTransposeRule(RuleQC()))
        big_right = SimplifiedBigSite(big_right_orig, NoTransposeRule(RuleQC()))
        hamil = HamiltonianQCBigSite(vacuum, n_orbs, VectorUInt8(orb_sym), fcidump,
            big_left, big_right)
        big_left_orig = CSFBigSite(n_inactive, self.ci_order, False,
            fd_dyall, VectorUInt8(orb_sym[:n_inactive]), max(min(self.verbose - 4, 3), 0))
        big_right_orig = CSFBigSite(n_external, self.ci_order, True,
            fd_dyall, VectorUInt8(orb_sym[-n_external:]), max(min(self.verbose - 4, 3), 0))
        big_left = SimplifiedBigSite(big_left_orig, RuleQC())
        big_right = SimplifiedBigSite(big_right_orig, RuleQC())
        hm_dyall = HamiltonianQCBigSite(vacuum, n_orbs, VectorUInt8(orb_sym), fd_dyall,
            big_left, big_right)

        # Left MPO
        lmpo = MPOQC(hm_dyall, QCTypes.NC)
        lmpo = SimplifiedMPO(lmpo, RuleQC(), True, True, OpNamesSet((OpNames.R, OpNames.RD)))
        lmpo.const_e -= self.e_tot
        lmpo = lmpo * -1

        # Right MPO
        rmpo = MPOQC(hamil, QCTypes.NC)
        rmpo = SimplifiedMPO(rmpo, NoTransposeRule(RuleQC()),
            True, True, OpNamesSet((OpNames.R, OpNames.RD)))
        rmpo.const_e -= self.e_tot

        # MPS
        mps.dot = 2
        if mps.center == mps.n_sites - 1 and mps.dot == 2:
            mps.center = mps.n_sites - 2
        bra_info = MPSInfo(n_sites, vacuum, target, hamil.basis)
        bra_info.tag = 'BRA'
        bra_info.set_bond_dimension(bond_dims[0])
        bra = MPS(n_sites, mps.center, 2)
        bra.initialize(bra_info)
        bra.random_canonicalize()
        bra.tensors[bra.center].normalize()
        bra.save_mutable()
        bra_info.save_mutable()
        if bra.center == 0 and bra.dot == 2:
            bra.move_left(hamil.opf.cg, None)
        elif bra.center == bra.n_sites - 2 and bra.dot == 2:
            bra.move_right(hamil.opf.cg, None)
        bra.center = mps.center

        # Linear
        lme = MovingEnvironment(lmpo, bra, bra, "LME")
        lme.init_environments(self.verbose >= 5)
        lme.delayed_contraction = OpNamesSet.normal_ops()
        lme.cached_contraction = False
        rme = MovingEnvironment(rmpo, bra, mps, "RME")
        rme.init_environments(self.verbose >= 5)
        linear = LinearBigSite(lme, rme, None, VectorUBond(bond_dims),
            VectorUBond([mps.info.bond_dim + 400]), VectorDouble(noises))
        linear.last_site_svd = True
        linear.last_site_1site = True
        linear.decomp_last_site = False
        linear.cutoff = self.dmrg_args["cutoff"]
        linear.iprint = max(min(self.verbose - 4, 3), 0)
        linear.linear_conv_thrds = VectorDouble([x / 50 for x in dav_thrds])

        self.e_corr = linear.solve(len(bond_dims), mps.center == 0, sweep_tol)

        lib.logger.note(self, 'E(NEVPT2) = %.16g  E_corr = %.16g', self.e_tot, self.e_corr)

        self.mps = bra

        if self.mp_order == 2:
            return self.e_corr, bra
        
    def make_rdm1(self, state=None, norb=None, nelec=None):
        '''
        Spin-traced one-particle density matrix in MO basis.
        dm1[p,q] = <q_alpha^\dagger p_alpha> + <q_beta^\dagger p_beta>
        '''

        if state is None:
            state = self.mps
            state.dot = 2
        
        if state.center == state.n_sites - 1 and state.canonical_form[state.center] == 'K':
            cg = CG(200)
            cg.initialize()
            state.move_left(cg, None)
            state.center = state.n_sites - 2
        
        # 1PDM MPO
        pmpo = PDM1MPOQC(self.hamil, 0)
        pmpo = SimplifiedMPO(pmpo, RuleQC())

        # 1PDM
        pme = MovingEnvironment(pmpo, state, state, "1PDM")
        pme.init_environments(self.verbose >= 5)
        expect = Expect(pme, state.info.bond_dim, state.info.bond_dim)
        expect.iprint = max(min(self.verbose - 4, 3), 0)
        expect.solve(True, state.center == 0)
        dmr = expect.get_1pdm_spatial(self.n_orbs)
        dm = np.array(dmr).copy()
        dmr.deallocate()

        return dm.transpose((1, 0))

def NEVPT2(*args, **kwargs):
    kwargs['mp_order'] = 2
    return NEVPT(*args, **kwargs)

if __name__ == '__main__':

    from pyscf import gto, scf, fci

    mol = gto.M(
        atom='''
            O  0.000000  0.000000  0.000000
            H  0.758602  0.000000  0.504284
            H  0.758602  0.000000  -0.504284
        ''',
        basis='6-31g',
        verbose=4, symmetry=False, spin=0)

    # Mean Field
    mf = scf.RHF(mol).run()

    # DMRG CASCI
    myci = mcscf.CASCI(mf, 8, 8)
    myci.fcisolver = _DMRGCI(mf)
    myci.fcisolver.dmrg_args['maxM'] = 1000
    myci.fcisolver.dmrg_args['sweep_tol'] = 1E-16
    myci = myci.run()
    dm1 = myci.make_rdm1()

    # DMRG NEVPT2
    mypt = NEVPT2(myci)
    mypt.dmrg_args['maxM'] = 1000
    mypt.dmrg_args['sweep_tol'] = 1E-16
    mypt = mypt.run()

    # DMRG NEVPT2 singles
    mypt = NEVPT2(myci, ci_order=1)
    mypt.dmrg_args['maxM'] = 1000
    mypt.dmrg_args['sweep_tol'] = 1E-16
    mypt = mypt.run()
